/*
** Module   : DynamicQueryHelper.java
** Abstract : Helper class for dynamic queries.
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130929 Created initial version.
** 002 SVL 20131115 Add "force_dmo_alias" and "is_dynamic_table" annotations for dynamic buffers.
** 003 CA  20131013 Added no-op deleted() method, required by the changes in Finalizable interface.
** 004 OM  20140122 Added support for parsing predicated and return of FindQuery -s needed by
**                  findFirst / findLast and findUnique methods of BufferImpl.
** 005 OM  20140128 Adding normalization when accessing TableMapper.getLegacyName().
** 006 SVL 20140320 loadPermanentSchemas loads schemas from resources. Removed unnecessary loading
**                  of registry.xml.
** 007 ECF 20140404 Reimplemented dynamic conversion to use ConversionPool.
** 008 OM  20140417 Added case-sensitive support.
** 009 ECF 20140425 Removed unused flag.
** 010 VMN 20140506 Added enhance schema name conversion support for hint "escape" attribute.
** 011 VMN 20140511 Fixed FIRST clause in preparePredicate.
** 012 SVL 20140601 Fixed emission of buffer name in H011.
** 013 VMN 20140708 Enhanced H002 fix: force_dmo_alias annotation is needed for both static and
**                  dynamic buffers, to ensure converted buffer names (DMO aliases) match the
**                  caller's DMO aliases.
** 014 ECF 20140918 Fixed memory leaks which were leaving ASTs behind in memory. Reduced footprint
**                  of P2OLookup instances by intern'ing string data.
** 015 VMN 20141121 Fixed dot and colon usages in predicate for QUERY-PREPARE() in order to
**                  handle 4GL behavior.
** 016 VMN 20141204 Fixed dot and colon usages when they are part of a string constant, embedded
**                  in the query predicate for QUERY-PREPARE().
** 017 ECF 20150105 Modified code to match SchemaDictionary.loadFromAst API signature change.
** 018 ECF 20150109 Normalize legacy field names when using them as hash map keys.
** 019 OM  20150116 Replaced the In-memory compilation with java JAST interpreter.
** 020 OM  20150217 Improved preparePredicate() for OPEN QUERY predicates.
** 021 OM  20150325 Stop processing query if ProgressParser reports syntax error.
** 022 ECF 20150328 Increased query cache size.
** 023 OM  20150307 Improved message printed on fail to process query string.
** 024 OM  20150429 Fixed debug output for the generated (intermediary and final) java code.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.*;
import java.util.logging.*;
import org.hibernate.mapping.*;
import antlr.*;
import com.goldencode.ast.*;
import com.goldencode.cache.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.persist.lock.LockType;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

import static com.goldencode.p2j.persist.DynamicTablesHelper.*;

/**
 * Helper class for dynamic query conversion.
 */
class DynamicQueryHelper
{
   /** Colon. */
   public static final char COLON = ':';
   
   /** Dot. */
   public static final char DOT = '.';
   
   /** Single quote. */
   public static final char QUOTE = '\'';
   
   /** Double quote. */
   public static final char DOUBLE_QUOTE = '"';
   
   /** Back slash. */
   public static final char BACK_SLASH = '\\';
   
   /** Tilde. */
   public static final char TILDE = '~';
   
   /**
    * Token to prepend to a find-first predicate in order to obtain a well formed P4GL query.
    * This is also used as a marker to later distinct whether the {@code pcode} is an FIND-FIRST
    * or OPEN QUERY statement.
    */
   private static final String FIND_FIRST_TOK = "FIND FIRST ";
   
   /** The name of the generated java class. */
   private static final String DYN_GEN_QUERY_CLASS = "DynGenQuery";
   
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(DynamicQueryHelper.class.getName());
   
   /** Map of P2J wrappers to their associated token type. */
   private static final Map<String, Integer> wrapperToTokenType = new HashMap<>();
   
   /** A counter of compiled dynamic query classes. */
   private static final AtomicLong queryCounter = new AtomicLong(1);
   
   /** Reference to context-local data. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   /**
    * Caches the queries using an aging & usage algorithm.
    * TODO: make the cache size configurable in the directory.
    */
   private static final ExpiryCache<QueryCacheKey, JavaAst> queryCache =
      new LFUAgingCache<>("DynamicQueryHelper", 4096);
   
   static
   {
      wrapperToTokenType.put(ParmType.CHAR.toString(),   ProgressParserTokenTypes.FIELD_CHAR);
      wrapperToTokenType.put(ParmType.INT.toString(),    ProgressParserTokenTypes.FIELD_INT);
      wrapperToTokenType.put(ParmType.INT64.toString(),  ProgressParserTokenTypes.FIELD_INT64);
      wrapperToTokenType.put(ParmType.DEC.toString(),    ProgressParserTokenTypes.FIELD_DEC);
      wrapperToTokenType.put(ParmType.LOG.toString(),    ProgressParserTokenTypes.FIELD_LOGICAL);
      wrapperToTokenType.put(ParmType.DATE.toString(),   ProgressParserTokenTypes.FIELD_DATE);
      wrapperToTokenType.put(ParmType.DT.toString(),     ProgressParserTokenTypes.FIELD_DATETIME);
      wrapperToTokenType.put(ParmType.DTTZ.toString(),
                             ProgressParserTokenTypes.FIELD_DATETIME_TZ);
      wrapperToTokenType.put(ParmType.HANDLE.toString(), ProgressParserTokenTypes.FIELD_HANDLE);
      wrapperToTokenType.put(ParmType.RECID.toString(),  ProgressParserTokenTypes.FIELD_RECID);
      wrapperToTokenType.put(ParmType.ROWID.toString(),  ProgressParserTokenTypes.FIELD_ROWID);
      
      // debug only: to see when the cache is full
      queryCache.addCacheExpiryListener(new CacheExpiryListener<QueryCacheKey, JavaAst>()
      {
         @Override
         public void cacheExpired(CacheExpiryEvent<QueryCacheKey, JavaAst> event)
         {
            if (LOG.isLoggable(Level.INFO))
            {
               LOG.log(Level.INFO, 
                       "DynamicQueryHelper cache is full.\n" +
                       "Starting to dump old entries: " + event.getSource());
            }
         }
      });
   }
   
   /**
    * Private worker for parsing a given query string using the QueryProcessor and return a
    * {@link P2JQuery} instance.
    * This method is only called by the package-public methods {@link #parseQuery} and
    * {@link #parseFindQuery}.
    *
    * @param   processor
    *          The processor that will generate the Progress code to be converted and later
    *          post-process the AST before generating the java code.
    * @param   buffers
    *          The list of buffers used by the query.
    * @param   predicate
    *          The query predicate.
    * @param   qname
    *          The query's name.
    *
    * @return  A {@link P2JQuery} instance representing the given legacy query string.
    */
   private static P2JQuery parse(QueryProcessor processor,
                                 List<Buffer> buffers,
                                 String predicate,
                                 String qname)
   {
      String        prefixedKey = getCacheKey(predicate, null);
      QueryCacheKey key         = new QueryCacheKey(prefixedKey, buffers);
      JavaAst       qast        = null;
      long          t0          = System.nanoTime(); // start the timer
      
      synchronized (queryCache)
      {
         // check the cache for a previously constructed JAST
         qast = queryCache.get(key);
      }
      
      if (qast == null)
      {
         // if not found, build it now
         qast = generateJavaTree(processor, buffers, predicate, qname);
         if (qast == null)
         {
            // any error message have already been logged
            return null;
         }
         
         synchronized (queryCache)
         {
            JavaAst racer = queryCache.get(key);
            if (racer == null)
            {
               // now we can store it for subsequent use
               queryCache.put(key, qast);
               
               if (LOG.isLoggable(Level.FINE))
               {
                  LOG.log(Level.FINE, "Added to cache as " + key + "\n" + qast.dumpTree());
               }
            }
            else
            {
               if (LOG.isLoggable(Level.INFO))
               {
                  LOG.log(Level.INFO,
                          "Query was processed in a parallel thread [" + predicate + "].");
               }
            }
         }
      }
      else
      {
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "Found in cache " + key + "\n" + qast.dumpTree());
         }
      }
      
      try
      {
         RuntimeJastInterpreter interpreter = new RuntimeJastInterpreter(buffers, null);
         interpreter.prepare(qast);                // load the jast tree
         interpreter.interpret("execute");         // interpret the execute() method
         return (P2JQuery) interpreter.getVariableValue(
               processor.getExpectedQueryName());  // extract the variable value
      }
      catch (Throwable t)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE,
                    "Failed to interpret JAST for " + key + "\n" + qast.dumpTree(), t);
         }
         return null;
      }
      finally
      {
         if (LOG.isLoggable(Level.FINE))
         {
            long dt = (System.nanoTime() - t0) / 1000; // drop precision to micro seconds
            LOG.log(Level.FINE, "DQH-Timer = " + dt / 1000.0 + " ms");
         }
      }
   }
   
   /**
    * Private worker for parsing a given query string using the QueryProcessor and return a
    * {@link JavaAst} instance.
    *
    * @param   processor
    *          The processor that will generate the Progress code to be converted and later
    *          post-process the AST before generating the java code.
    * @param   buffers
    *          The list of buffers used by the query.
    * @param   predicate
    *          The query predicate.
    * @param   qname
    *          The query's name.
    *
    * @return  a {@link JavaAst} instance representing a tree for a java compile unit.
    */
   private static JavaAst generateJavaTree(QueryProcessor processor,
                                           List<Buffer> buffers,
                                           String predicate,
                                           String qname) 
   {
      WorkArea wa = locate();
      long queryId = queryCounter.getAndIncrement();
      
      AstManager astManager = AstManager.get();
      
      // set the info about what we are currently parsing
      wa.predicate     = predicate;
      wa.qname         = qname;
      wa.buffers       = buffers;
      wa.queryId       = queryId;
      wa.dynQueryFile  = String.format(".%sin-mem-p2j-query%d.p", File.separator, queryId);
      wa.jastFile      = String.format("%s.jast", wa.dynQueryFile);
      wa.astFile       = String.format("%s.ast", wa.dynQueryFile);
      wa.p2oFile       = String.format("%s.p2o", wa.dynQueryFile);
      
      try
      {
         StringBuilder    pcode = new StringBuilder();
         SymbolResolver   sym   = new SymbolResolver(); // build the symbol resolver
         SchemaDictionary dict  = SchemaDictionary.localInstance(); 
         sym.setSchemaDictionary(dict); // set the schema dictionary used by this context
         
         try
         {
            // step 2: load p2o and schema definitions for all temporary buffers
            
            // add a default scope that represents the scope of the external procedure
            sym.addSchemaScope(false);
            
            try
            {
               prepareTempTable(dict, pcode); // load the p2o and schema data
            }
            catch (PersistenceException exc)
            {
               showConversionError(exc, "prepare the temp-table buffers for", pcode.toString());
               
               if (LOG.isLoggable(Level.SEVERE))
               {
                  String msg = "Error preparing the temp-table buffers for predicate '%s'";
                  LOG.log(Level.SEVERE, String.format(msg, predicate));
               }
               
               return null;
            }
            
            // step 3: parse the predicate
            predicate = processor.preparePredicate(predicate);
            if (predicate == null)
            {
               return null;
            }
            pcode.append(predicate);
            
            boolean        complete = false;
            Exception      err      = null;
            String         oeCode   = pcode.toString();
            StringReader   in       = new StringReader(oeCode);
            ProgressLexer  lexer    = new ProgressLexer(in, sym);
            ProgressParser parser   = new ProgressParser(lexer, sym);
            // debug: System.out.println("Starting dynamic conversion for P4GL code:\n" + oeCode);
            
            try
            {
               parser.setConsumeError(false); // throw exception instead of print event to console
               complete = parser.external_proc();
            }
            catch (RecognitionException | TokenStreamException exc)
            {
               err = exc;
               complete = false;
            }
            catch (RuntimeException exc2)
            {
               // if the RecognitionException was re-thrown wrapped as RuntimeException
               // from ProgressParser.reportError() we need to get the original back
               if (exc2.getCause() instanceof RecognitionException)
               {
                  err = (RecognitionException) exc2.getCause();
                  complete = false;
               }
               else
               {
                  // rethrow the exception of unknown cause
                  throw exc2;
               }
            }
            
            if (!complete)
            {
               showConversionError(err, "parse", oeCode);
               return null;
            }
            
            try
            {
               ProgressAst progAst = (ProgressAst) parser.getAST();
               progAst.brainwash(wa.dynQueryFile, true);
               
               // step 4: prepare the initial AST
               prepareQueryTree(progAst);
               
               // step 5: run conversion
               ConversionPool.runTask(ConversionProfile.ANNOTATIONS, progAst);
               progAst = (ProgressAst) astManager.loadTree(wa.astFile);
               
               ConversionPool.runTask(ConversionProfile.BASE_STRUCTURE, progAst);
               progAst = (ProgressAst) astManager.loadTree(wa.astFile);
               
               ConversionPool.runTask(ConversionProfile.CORE_CONVERSION, progAst);
               progAst = (ProgressAst) astManager.loadTree(wa.astFile);
               
               // walk the AST and collect all the buffers
               List<String> realBuffers = collectBuffers(progAst);
               
               // validate the buffers
               if (!validateBuffers(realBuffers))
               {
                  return null;
               }
               
               JavaAst jCode = (JavaAst) astManager.loadTree(wa.jastFile);
               // debug: System.out.println(jCode.dumpTree(true));
               // debug: System.out.println(ConversionPool.runTask(ConversionProfile.BREW, jCode)
               //                      .getStoredObject(DYN_GEN_QUERY_CLASS));
               processor.postprocessJavaAst(jCode);
               
               JavaAst qast = (JavaAst) astManager.loadTree(wa.jastFile);
               // debug: System.out.println(qast.dumpTree(true));
               // debug: System.out.println(ConversionPool.runTask(ConversionProfile.BREW, qast)
               //              .getStoredObject(DYN_GEN_QUERY_CLASS));
               return qast;
            }
            catch (Exception exc)
            {
               showConversionError(exc, "convert", oeCode);
               return null;
            }
         }
         finally
         {
            // clean up the schema
            sym.deleteSchemaScope(false);
            
            // clean up the progress, java, and P2O ASTs, they are no longer needed
            astManager.removeTree(wa.dynQueryFile);
            astManager.removeTree(wa.astFile);
            astManager.removeTree(wa.jastFile);
            astManager.removeTree(wa.p2oFile);
         }
      }
      finally
      {
         // reset the WorkArea temporary data
         wa.predicate     = null;
         wa.qname         = null;
         wa.queryId       = 0;
         wa.buffers       = null;
         wa.dynQueryFile  = null;
         wa.jastFile      = null;
         wa.astFile       = null;
         wa.p2oFile       = null;
      }
   }
   
   /**
    * Computes a string that will be used in the cache key. For the moment, the original pred
    * string is used (UPPERCASED) to implement a naive caching (that matches char-to char query).
    * Future algorithms will probably use the progress ast to match similar queries and extract
    * string (and possible other data types: int, boolean) constants parameters.
    * 
    * @param   pred
    *          The original pred.
    * @param   progAst
    *          The AST tree that was generated on that tree. Not used at this moment.
    * 
    * @return  A string that will identify the predicate in the cache (uppercased ?).
    */
   private static String getCacheKey(String pred, ProgressAst progAst)
   {
      return pred; // .toUpperCase();
   }
   
   /**
    * Parse the given query string and return a {@link P2JQuery} instance.
    * 
    * @param    buffers
    *           The list of buffers used by the query.
    * @param    predicate
    *           The query predicate.
    * @param    qname
    *           The query's name.
    * 
    * @return   A {@link P2JQuery} instance representing the given legacy query string.
    */
   static P2JQuery parseQuery(List<Buffer> buffers, String predicate, String qname)
   {
      QueryProcessor openQueryProcessor = new QueryProcessor()
      {
         @Override
         public String getExpectedQueryName()
         {
            return "query0";
         }
         
         /**
          * Creates a Progress syntax correct <code>OPEN QUERY</code> statement that uses the
          * given predicate. It will be later converted and compiled.
          * <p> 
          * <strong>Note</strong>: This is not a perfect check for query sanity, but it will:
          * <ul>
          *    <li>allow one level nested quotes
          *    <li>allow handle methods as valid constructs: {@code buffer book:name}
          *    <li>allow dereference operator (:: punctuation): {@code buffer book::isbn}
          *    <li>drop garbage after the COLON (:) or DOT (.) when they are terminators 
          *          (followed by some white space) 
          * </ul>
          *
          * @param   predicate
          *          The query predicate that needs to be dynamically executed.
          *
          * @return  A string with a Progress query statement that uses the predicate.
          */
         @Override
         public String preparePredicate(String predicate)
         {
            int len = predicate.length();
            int endIndex = -1;
            char quote = '\0';
            
            for (int i = 0; i < len; i++)
            {
               char current = predicate.charAt(i);
               if (current == QUOTE || current == DOUBLE_QUOTE)
               {
                  if (quote != 0 && quote != current)
                  {
                     continue; // nested quotes, not the correct pair!
                  }
                  if (i == 0)
                  {
                     // actually, will never happen because predicates do not start with
                     // character literals 
                     quote = current;
                  }
                  else
                  {
                     char prev = predicate.charAt(i - 1);
                     if (prev != TILDE && prev != BACK_SLASH)
                     // TODO: check the OS!, on Win32, BACK_SLASH is not normally an escape char
                     {
                        // if found the matching quote, mark as not inside the literal
                        // otherwise (a CHARACTER literal starts) store the current quote type
                        quote = (quote == current) ? '\0' : current;
                     }
                  }
               }
               else if (quote == '\0')
               {
                  if (current == DOT || current == COLON)
                  {
                     if (i == len - 1)
                     {
                        // do nothing, current DOT or COLON used as terminator
                        endIndex = i;
                        break; // will break anyway
                     }
                     else
                     {
                        char next = predicate.charAt(i + 1);
                        if (next == ' ' || next == '\t' || next == '\n' || next == '\r')
                        {
                           // current DOT or COLON used as terminator, drop following garbage
                           endIndex = i;
                           break;
                        } 
                     }
                     // else the text following DOT or COLON should be the name of a field,
                     // method or dereference. We don't know at this time, the Progress parser
                     // must be run to detect syntactic errors.
                  }
               }
            }
            
            if (endIndex != -1)
            {
               // remove garbage if detected
               predicate = predicate.substring(0, endIndex);
            }
            
            return "OPEN QUERY " + DYN_GEN_QUERY_CLASS + " " + predicate;
         }
         
         /**
          * Post-process the intermediary AST to clean it up from unneeded nodes that were
          * automatically created during the progress conversion and add accessor for the query.
          *
          * @param   jcode
          *          The intermediary jast as it was generated form prepared predicate.
          *
          * @throws  ConfigurationException
          *          if any error occurs loading the specified configuration profile.
          * @throws  AstException
          *          if any error occurs loading a persisted AST.
          */
         @Override
         public void postprocessJavaAst(JavaAst jcode)
         throws ConfigurationException,
                AstException
         {
            ConversionPool.runTask(ConversionProfile.POSTPROCESS_OPEN_QUERY, jcode);
         }
      };
      
      return parse(openQueryProcessor, buffers, predicate, qname);
   }
   
   /**
    * Parse the given predicate string and return a {@link FindQuery} instance.
    *
    * @param   buffer
    *          The buffer used by the query.
    * @param   predicate
    *          The predicate to be parsed.
    * @param   lock
    *          The locking mode, may include the NO-WAIT flag.
    * @param   qname
    *          The query's name.
    *
    * @return  A {@link P2JQuery} instance representing the given legacy query string.
    */
   static P2JQuery parseFindQuery(Buffer buffer, String predicate, LockType lock, String qname)
   {
      WorkArea wa = locate();
      try
      {
         // allow processing unknown predicates
         predicate = (predicate == null) ? "" : predicate.trim();
         
         final ArrayList<Buffer> buffers = new ArrayList<>();
         buffers.add(buffer);
         
         // save to context local for later use by findQueryProcessor
         wa.lock = lock;
         
         QueryProcessor findQueryProcessor = new QueryProcessor()
         {
            @Override
            public String getExpectedQueryName()
            {
               return "findQuery";
            }
            
            /**
             * Creates a Progress syntax correct statement that uses the given predicate.
             * It will be later converted and compiled. This method uses the context local
             * variable lock in order to achieve the required lock / wait behavior.
             *
             * @param   predicate
             *          Predicate expression for find that evaluates to the following syntax:
             *             [ WHERE [ logical-expression ] ] [ USE-INDEX index-name ]
             *
             * @return  A string with a Progress <code>FIND FIRST</code> query statement that
             *          uses the predicate.
             */
            @Override
            public String preparePredicate(String predicate)
            {
               WorkArea wa = locate();
               if (wa.buffers == null || wa.buffers.isEmpty())
               {
                  // this should never happen, anyway
                  return null;
               }
               
               // we introduce a FIRST clause that will be later removed in the postprocessing
               String pCode = FIND_FIRST_TOK + wa.buffers.get(0).name().toStringMessage() + " " +
                              predicate;
               // then add the required access locking
               if (wa.lock.isExclusive())
               {
                  pCode += " EXCLUSIVE-LOCK";
               }
               else if (!wa.lock.isShare())
               {
                  pCode += " NO-LOCK";
               }
               else
               {
                  pCode += " SHARE-LOCK"; // default
               }
               
               if (wa.lock.isNoWait())
               {
                  pCode += " NO-WAIT";
               }
               
               return pCode + ".";
            }
            
            /**
             * Post-process the intermediary AST to clean it up from unneeded nodes that were
             * automatically created during the progress conversion and add accessor for
             * the query.
             *
             * @param   jcode
             *          The intermediary jast as it was generated form prepared predicate.
             * 
             * @throws  ConfigurationException
             *          if any error occurs loading the specified configuration profile.
             * @throws  AstException
             *          if any error occurs loading a persisted AST.
             */
            @Override
            public void postprocessJavaAst(JavaAst jcode)
            throws ConfigurationException,
                   AstException
            {
               ConversionPool.runTask(ConversionProfile.POSTPROCESS_FIND_QUERY, jcode);
            }
         };
         
         return parse(findQueryProcessor, buffers, predicate, qname);
      }
      finally
      {
         // cleanup
         wa.lock = null;
      }
   }
   
   /**
    * Process the definitions of the temp-tables (associated with the temporary buffers in the
    * given list) to create the associated dictionary and p2o ASTs.
    * <p>
    * Also, it appends to the progress code string any needed 4GL code (like DEFINE BUFFER 
    * statements for explicit buffers, to refer back the originating table.
    * 
    * @param    dict
    *           The schema dictionary where the temp-tables will be loaded.
    * @param    pcode
    *           Append any new progress code here.
    * 
    * @throws   PersistenceException
    *           In case of problems during preparation of the temp-table schema and p2o ASTs.
    */
   private static void prepareTempTable(SchemaDictionary dict, StringBuilder pcode)
   throws PersistenceException
   {
      WorkArea wa = locate();
      
      DataModelAst p2oRoot = buildRootDataModel(DatabaseManager.TEMP_TABLE_SCHEMA);
      p2oRoot.brainwash(wa.p2oFile);
      ProgressAst schemaRoot = buildRootSchema(DatabaseManager.TEMP_TABLE_SCHEMA);
      
      integer tableOrder = new integer(0);
      for (Buffer buf : wa.buffers)
      {
         BufferImpl impl = (BufferImpl) buf;
         RecordBuffer rbuf = impl.buffer();
         rbuf.initialize();
         
         String legacyBufName = impl.name().toStringMessage();
         String legacyTableName = TableMapper.getLegacyName(buf);
         
         if (rbuf.isTemporary())
         {
            // create the schema and p2o definition for this temp-table
            DataModelAst model = buildDataModelClass(buf, rbuf);
            ProgressAst table = buildTempTable(model, rbuf, tableOrder);
            
            schemaRoot.graft(table);
            
            p2oRoot.graft(model);
         }
         
         // must create explicit buffers, if the legacy buffer name is not the same as the legacy
         // default buffer name (which is the same as the legacy table name).
         if (!legacyBufName.equalsIgnoreCase(legacyTableName))
         {
            String tempTable = rbuf.isTemporary() ? "TEMP-TABLE" : ""; 
            
            String defBufStmt = String.format("DEFINE BUFFER %s FOR %s %s.%n", 
                                              legacyBufName, tempTable, legacyTableName);
            pcode.append(defBufStmt);
         }
      }
      
      // save the P2O tree, to expose it to the conversion rules
      AstManager.get().saveTree(p2oRoot, wa.p2oFile, true);
      
      // load the temp schema
      dict.loadFromAst(schemaRoot, false);
   }
   
   /**
    * Prepare the query AST by placing the appropriate annotations with the converted java names
    * for the external program and also by created bogus
    * {@link ProgressParserTokenTypes#DEFINE_TEMP_TABLE} nodes (which would be created from a
    * <code>DEFINE TEMP-TABLE tt.</code> statement, with no fields); the later is needed to allow
    * proper conversion of the default buffer name.
    * 
    * @param    root
    *           The root AST of the converted progress code.
    */
   private static void prepareQueryTree(AnnotatedAst root)
   {
      WorkArea wa = locate();
      
      // Create a bogus DEFINE TEMP-TABLE for each temp buffer (just to let their converted
      // buffer names resolve properly).
      // Dynamic buffers have unique DMO aliases, so "some-buffer" name may be converted to the
      // alias like "someTable__1", while aliases in predicate will be converted using standard
      // P2J rules, which will result to the converted alias "someTable". So below we explicitly
      // specify converted DMO alias like "someTable__1" for the conversion rules using
      // "force_dmo_alias" annotation.
      for (Buffer buf : wa.buffers)
      {
         BufferImpl impl = (BufferImpl) buf;
         String legacyTableName = DynamicTablesHelper.normalizeName(
               TableMapper.getLegacyName(buf));
         
         if (!impl.buffer().isTemporary())
         {
            continue;
         }
         
         ProgressAst stmt = createProgressAst(ProgressParserTokenTypes.STATEMENT, "statement");
         stmt.putAnnotation("reachable", true);
         
         ProgressAst defChild = createProgressAst(ProgressParserTokenTypes.DEFINE_TEMP_TABLE,
                                                  "DEFINE");
         defChild.putAnnotation("reachable", true);
         defChild.putAnnotation("schemaname", legacyTableName);
         defChild.putAnnotation("bufname", legacyTableName);
         defChild.putAnnotation("dbname", "");
         if (impl.buffer().isDynamic())
         {
            defChild.putAnnotation("force_dmo_alias", impl.buffer().getDMOAlias());
            if (DynamicTablesHelper.isDynamicTableBuffer(impl.buffer()))
            {
               defChild.putAnnotation("is_dynamic_table", true);
            }
         }
         stmt.graft(defChild);
         
         ProgressAst ttChild = createProgressAst(ProgressParserTokenTypes.KW_TEMP_TAB,
                                                 "TEMP-TABLE");
         ttChild.putAnnotation("reachable", true);
         defChild.graft(ttChild);
         
         ProgressAst symChild = createProgressAst(ProgressParserTokenTypes.SYMBOL,
                                                  legacyTableName);
         symChild.putAnnotation("reachable", true);
         ttChild.graft(symChild);
         
         root.graftAt(stmt, 0);
      }
      
      // put annotation related to name, package, etc
      root.putAnnotation("file", wa.dynQueryFile);
      root.putAnnotation("classname", DYN_GEN_QUERY_CLASS);
      root.putAnnotation("javaname", DYN_GEN_QUERY_CLASS);
      root.putAnnotation("relative-name", wa.dynQueryFile);
      root.putAnnotation("basename", wa.dynQueryFile);
      root.putAnnotation("package-base", "");
      root.putAnnotation("pkgname", "com.goldencode.p2j.persist.dynquery");
      
      // make "force_dmo_alias" annotations, errors are ignored at this point, they will
      // be raised later anyway
      Iterator<Aast> iter = root.iterator();
      while (iter.hasNext())
      {
         Aast chAst = iter.next();
         
         if (chAst.getType() == ProgressParserTokenTypes.RECORD_PHRASE ||
             chAst.getType() == ProgressParserTokenTypes.DEFINE_BUFFER)
         {
            Aast bufAst = (Aast) chAst.getFirstChild();
            String bufName = (String) bufAst.getAnnotation("bufname");
            if (bufName != null)
            {
               int dotPos = bufName.lastIndexOf(".");
               if (dotPos >= 0)
               {
                  bufName = bufName.substring(dotPos + 1);
               }
               
               int idx = locateBuffer(bufName);
               if (idx >= 0)
               {
                  BufferImpl impl = (BufferImpl) wa.buffers.get(idx);
                  bufAst.putAnnotation("force_dmo_alias", impl.buffer().getDMOAlias());
                  if (DynamicTablesHelper.isDynamicTableBuffer(impl.buffer()))
                  {
                     bufAst.putAnnotation("is_dynamic_table", true);
                  }
               }
            }
         }
      }
   }
   
   /**
    * Get the buffer names used by the AST produced for converted predicated.
    * 
    * @param    result
    *           The list of buffer names.
    *           
    * @return   See above.
    */
   private static List<String> collectBuffers(ProgressAst result)
   {
      List<String> buffers = new ArrayList<>();
      
      Iterator<Aast> iter = result.iterator();
      while (iter.hasNext())
      {
         Aast chAst = iter.next();
         if (chAst.getType() != ProgressParserTokenTypes.RECORD_PHRASE)
         {
            continue;
         }
         
         Aast bufAst = (Aast) chAst.getFirstChild();
         String bufname = (String) bufAst.getAnnotation("bufname");
         if (bufname.contains("."))
         {
            bufname = bufname.substring(bufname.lastIndexOf(".") + 1);
         }
         
         buffers.add(bufname);
      }
      
      return buffers;
   }
   
   /**
    * Validate the list of buffers against the buffers loaded by this QUERY resource (loaded in
    * the {@link WorkArea#buffers} list.
    * <p>
    * The buffers must have (all conditions must be true):
    * <ol>
    *    <li>The buffers must be in the same order as at the QUERY resource.</li>
    *    <li>All the buffers loaded by this QUERY resource must be referenced.</li>
    *    <li>No buffer outside the QUERY's buffer list is allowed.</li>
    * </ol>
    * 
    * @param    realBuffers
    *           The list of buffer names used at the query predicated.
    *           
    * @return   <code>true</code> if the buffers are valid.
    */
   private static boolean validateBuffers(List<String> realBuffers)
   {
      WorkArea wa = locate();
      
      // check if the expected and converted buffer list match
      if (realBuffers.size() != wa.buffers.size())
      {
         final String errMsg =
            "QUERY-PREPARE text must have 1 FOR EACH/PRESELECT for each query buffer";
         
         ErrorManager.recordOrShowError(7325, errMsg, true, false, false);
         
         return false;
      }
      
      // check buffer positions and if known buffer
      for (int i = 0; i < realBuffers.size(); i++)
      {
         String realBuffer = realBuffers.get(i);
         
         int idx = locateBuffer(realBuffer);
         if (idx == -1)
         {
            String errMsg = " %s must be an unabbreviated name of a buffer known in query %s";
            errMsg = String.format(errMsg, realBuffer, wa.qname);
            
            ErrorManager.recordOrShowError(7327, errMsg, true, false, false);
            
            return false;
         }
         
         if (idx != i)
         {
            String errMsg = "Buffer %s is not referenced in the same order in PREPARE as " +
                            "in BUFFERS list for query %s";
            errMsg = String.format(errMsg, realBuffer, wa.qname);
            
            ErrorManager.recordOrShowError(7326, errMsg, true, false, false);
            
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Locate a {@link Buffer} with the given <code>name</code>, from the {@link WorkArea#buffers}
    * list and return its position.
    * 
    * @param    name
    *           The buffer name.
    *           
    * @return   The 0-based position of this buffer or -1 if not found.
    */
   private static int locateBuffer(String name)
   {
      WorkArea wa = locate();
      
      int idx = 0;
      for (Buffer buf : wa.buffers)
      {
         if (name.equalsIgnoreCase(buf.name().toStringMessage()))
         {
            return idx;
         }
         
         idx++;
      }
      
      return -1;
   }
   
   /**
    * Build a temp-table definition to be loaded in the {@link SchemaDictionary}. Use the given
    * {@link DataModelAst data model object} AST as the source.
    * 
    * @param    p2o
    *           The data model object AST.
    * @param    rbuf
    *           The record buffer instance.
    * @param    order
    *           The order of this temp-table AST.
    *           
    * @return   See above.
    */
   private static ProgressAst buildTempTable(DataModelAst p2o, RecordBuffer rbuf, integer order)
   {
      String schema      = rbuf.getSchema();
      Class<?> iface     = rbuf.getDMOInterface();
      String legacyTable = (String) p2o.getAnnotation("historical");
      
      ProgressAst root = createProgressAst(SchemaParserTokenTypes.TEMP_TABLE, legacyTable);
      root.putAnnotation("lastOrder", order.toJavaLongType());
      
      int fieldIdx = 0;
      for (int i = 0; i < p2o.getNumImmediateChildren(); i++)
      {
         DataModelAst child = (DataModelAst) p2o.getChildAt(i);
         
         if (child.getType() == DataModelTokenTypes.PROPERTY)
         {
            String datatype = (String) child.getAnnotation("datatype");
            int ftype = wrapperToTokenType.get(datatype);
            String legacyField = (String) child.getAnnotation("historical");
            
            ProgressAst field = createProgressAst(ftype, legacyField);
            
            // the following properties are not needed by SchemaDictionary:
            // 1. sqlWidth
            // 2. extent
            // 3. decimals
            // 4. length
            // 5. maxwidth
            
            if (DMOIndex.isCaseSensitive(schema, iface, child.getText()))
            {
               ProgressAst pcase = createProgressAst(SchemaParserTokenTypes.KW_CASE_SEN,
                                                     "case-sensitive");
               pcase.addChild(createProgressAst(SchemaParserTokenTypes.BOOL_TRUE, "true"));
               field.addChild(pcase);
            }
            
            // add ORDER node
            ProgressAst pord = createProgressAst(SchemaParserTokenTypes.ORDER, "ORDER");
            pord.addChild(createProgressAst(SchemaParserTokenTypes.NUM_LITERAL,
                                            Long.toString(fieldIdx)));
            fieldIdx = fieldIdx + 10;
            
            field.addChild(pord);
            
            Long extent = (Long) child.getAnnotation("extent");
            if (extent != null)
            {
               ProgressAst pext = createProgressAst(SchemaParserTokenTypes.KW_EXTENT, "extent");
               pext.addChild(createProgressAst(SchemaParserTokenTypes.NUM_LITERAL, 
                                               extent.toString()));
               
               field.addChild(pext);
            }
            
            root.addChild(field);
         }
         else if (child.getType() == DataModelTokenTypes.INDEX)
         {
            // add a new index
            String legacyIndex = (String) child.getAnnotation("historical");
            ProgressAst rootIdx = createProgressAst(SchemaParserTokenTypes.INDEX, legacyIndex);
            
            // build index properties
            ProgressAst props = createProgressAst(SchemaParserTokenTypes.PROPERTIES, "");
            if (child.isAnnotation(P2OLookup.UNIQUE))
            {
               props.addChild(createProgressAst(SchemaParserTokenTypes.KW_UNIQUE, "UNIQUE"));
            }
            if (child.isAnnotation(P2OLookup.PRIMARY))
            {
               props.addChild(createProgressAst(SchemaParserTokenTypes.KW_PRIMARY, "PRIMARY"));
            }
            if (child.isAnnotation(P2OLookup.WORD))
            {
               props.addChild(createProgressAst(SchemaParserTokenTypes.KW_WORD_IDX, "WORD"));
            }
            
            rootIdx.addChild(props);
            
            // build index components
            DataModelAst idxCol =
               (DataModelAst) child.getImmediateChild(DataModelTokenTypes.INDEX_COL, null);
            while (idxCol != null)
            {
               String legacyField = (String) child.getAnnotation("historical");
               
               ProgressAst idxField = createProgressAst(SchemaParserTokenTypes.INDEX_FIELD,
                                                        legacyField);
               if (idxCol.isAnnotation(P2OLookup.ABBREVIATED))
               {
                  idxField.addChild(createProgressAst(SchemaParserTokenTypes.KW_ABBV,
                                                      "ABBREVIATED"));
               }
               if (idxCol.isAnnotation(P2OLookup.DESCEND))
               {
                  idxField.addChild(createProgressAst(SchemaParserTokenTypes.KW_DESCEND,
                                                      "DESCENDING"));
               }
               else
               {
                  idxField.addChild(createProgressAst(SchemaParserTokenTypes.KW_ASCEND,
                                                      "ASCENDING"));
               }
               
               rootIdx.addChild(idxField);
               
               idxCol = (DataModelAst) child.getImmediateChild(DataModelTokenTypes.INDEX_COL,
                                                               idxCol);
            }
            
            root.addChild(rootIdx);
         }
      }
      
      order.assign(order.intValue() + 10);
      
      return root;
   }
   
   /**
    * Build the root {@link ProgressAst} node for the schema dictionary AST.
    * 
    * @param    schema
    *           The schema name.
    *           
    * @return   See above.
    */
   private static ProgressAst buildRootSchema(String schema)
   {
      ProgressAst root = createProgressAst(SchemaParserTokenTypes.DATABASE, schema);
      root.putAnnotation("srcfile", "in-memory");
      
      return root;
   }
   
   /**
    * Build the {@link DataModelAst root} AST and associated with the given schema.
    * 
    * @param    schema
    *           The schema name.
    *           
    * @return   See above.
    */
   private static DataModelAst buildRootDataModel(String schema)
   {
      String pkg = String.format("%s.dmo", Configuration.getParameter("pkgroot"));
      
      // this is a reverse generation of what schema/p2o.xml does
      DataModelAst root = createDataModelAst(DataModelTokenTypes.DATA_MODEL, schema);
      root.putAnnotation(P2OLookup.PACKAGE, pkg);
      
      return root;
   }
   
   /**
    * Create a {@link DataModelAst data model object} AST for the specified buffer.
    * 
    * @param    buf
    *           The buffer instance.
    * @param    rbuf
    *           The {@link RecordBuffer} instance for the buffer.
    * 
    * @return   A {@link DataModelAst} buffer definition.
    * 
    * @throws   PersistenceException
    *           In case of problems during preparation of the temp-table schema and p2o ASTs.
    */
   private static DataModelAst buildDataModelClass(Buffer buf, RecordBuffer rbuf)
   throws PersistenceException
   {
      String historical =
         DynamicTablesHelper.normalizeName(TableMapper.getLegacyName(buf)).intern();
      String dmoName = rbuf.getDMOName().intern();
      
      DataModelAst root = createDataModelAst(DataModelTokenTypes.CLASS, dmoName);
      root.putAnnotation(P2OLookup.HISTORICAL, historical);
      
      // build the ID property
      DataModelAst idProp = createDataModelAst(DataModelTokenTypes.PRIMARY, "id");
      idProp.putAnnotation(P2OLookup.DATATYPE, "Long");
      root.addChild(idProp);
      
      Map<String, DataModelAst> propMap = new HashMap<>();
      
      // go through all non-extent properties
      Class<?> dmoClass = rbuf.getDMOImplementationClass();
      Field[] fields = dmoClass.getDeclaredFields();
      for (Field field : fields)
      {
         if (!BaseDataType.class.isAssignableFrom(field.getType()))
         {
            continue;
         }
         
         DataModelAst propAst = buildProperty(buf, field, 0);
         
         root.addChild(propAst);
         
         String histName = (String) propAst.getAnnotation(P2OLookup.HISTORICAL);
         propMap.put(normalizeName(histName), propAst);
      }
      
      // add the extent properties
      Class<?>[] innerClasses = dmoClass.getDeclaredClasses();
      for (Class<?> inner : innerClasses)
      {
         int extent = 0;
         
         fields = inner.getDeclaredFields();
         for (Field field : fields)
         {
            if (!BaseDataType.class.isAssignableFrom(field.getType()))
               continue;
            
            if (extent == 0)
            {
               String fname = field.getName();
               
               String sizerMthdName = String.format("size%s%s",
                                                    Character.toUpperCase(fname.charAt(0)),
                                                    fname.substring(1));
               try
               {
                  Method sizerMthd = dmoClass.getMethod(sizerMthdName);
                  // We cannot call sizer method on a buffer because it may contain no record,
                  // so we create an unattached DMO and call sizer method on it.
                  Constructor ctor = dmoClass.getConstructor();
                  Object dmo = ctor.newInstance();
                  extent = ((integer) sizerMthd.invoke(dmo)).toJavaIntegerType();
               }
               catch (InstantiationException    |
                      IllegalAccessException    |
                      IllegalArgumentException  |
                      InvocationTargetException |
                      NoSuchMethodException     |
                      SecurityException exc)
               {
                  String msg = "Error invoking the sizer method %s for property %s";
                  msg = String.format(msg, sizerMthdName, fname);
                  
                  throw new PersistenceException(msg, exc);
               }
            }
            
            root.addChild(buildProperty(buf, field, extent));
         }
      }
      
      // composite info is not needed by P2OLookup
      
      Iterator<P2JIndex> idxIter = DMOIndex.databaseIndexes(rbuf.getSchema(),
                                                            rbuf.getDatabase(),
                                                            rbuf.getDMOInterface());
      while (idxIter.hasNext())
      {
         P2JIndex index = idxIter.next();
         String idx = DBUtils.extractIndexName(index.getName());
         
         String legacyName = TableMapper.getLegacyIndexName(buf, idx);
         if (legacyName != null)
         {
            // include only legacy indexes
            root.addChild(buildIndex(buf, rbuf, index, idx, propMap));
         }
      }
      
      // DataModelTokenTypes.UNIQUE, ONE_TO_ONE and ONE_TO_MANY nodes are not needed
      
      return root;
   }
   
   /**
    * Create a {@link DataModelAst property} AST for the specified DMO java field.
    * 
    * @param    buf
    *           The {@link Buffer} instance to which the property belongs.
    * @param    field
    *           The java field defining this property at the DMO.
    * @param    extent
    *           The legacy extent of this property.
    *           
    * @return   A {@link DataModelAst} property definition.
    */
   private static DataModelAst buildProperty(Buffer buf, Field field, int extent)
   {
      String property = field.getName().intern();
      RecordBuffer recordBuffer = ((BufferImpl) buf).buffer();
      String historical;
      if (DynamicTablesHelper.isDynamicTableBuffer(recordBuffer))
      {
         historical = DynamicTablesHelper.get4GLFieldName(recordBuffer.getDMOInterface(),
                                                          property);
      }
      else
      {
         historical = TableMapper.getLegacyFieldName(buf, property);
      }
      String datatype = field.getType().getName();
      datatype = datatype.substring(datatype.lastIndexOf('.') + 1).intern();
      
      DataModelAst root = createDataModelAst(DataModelTokenTypes.PROPERTY, property);
      root.putAnnotation(P2OLookup.HISTORICAL, historical.intern());
      root.putAnnotation(P2OLookup.DATATYPE, datatype);
      if (extent != 0)
      {
         root.putAnnotation(P2OLookup.EXTENT, new Long(extent));
      }
      if ("character".equals(datatype))
      {
         LegacyField legacy = field.getAnnotation(LegacyField.class);
         if (legacy.caseSensitive())
         {
            root.putAnnotation(P2OLookup.CASE_SENSITIVE, true);
         }
      }
      
      // INITIAL is not needed for P2OLookup
      
      return root;
   }
   
   /**
    * Build a {@link DataModelAst index} for the specified buffer.
    * 
    * @param    buf
    *           The {@link Buffer} instance to which the index belongs.
    * @param    rbuf
    *           The {@link RecordBuffer} instance.
    * @param    index
    *           The index definition.
    * @param    idxName
    *           The index name.
    * @param    propMap
    *           A map with the {@link DataModelAst property} ASTs by their legacy name.
    * 
    * @return   A {@link DataModelAst index} definition.
    */
   private static DataModelAst buildIndex(Buffer                    buf,
                                          RecordBuffer              rbuf,
                                          P2JIndex                  index,
                                          String                    idxName,
                                          Map<String, DataModelAst> propMap)
   {
      String historical = TableMapper.getLegacyIndexName(buf, idxName).intern();
      boolean unique    = index.isUnique();
      boolean primary   = index.isPrimary();
      boolean word      = index.isWord();
      
      DataModelAst root = createDataModelAst(DataModelTokenTypes.INDEX, idxName);
      root.putAnnotation(P2OLookup.HISTORICAL, historical);
      if (unique)
      {
         root.putAnnotation(P2OLookup.UNIQUE, unique);
      }
      if (primary)
      {
         root.putAnnotation(P2OLookup.PRIMARY, primary);
      }
      if (word)
      {
         root.putAnnotation(P2OLookup.WORD, word);
      }
      
      PersistentClass pclass = DatabaseManager.getClassMapping(rbuf.getDatabase(),
                                                               rbuf.getDMOImplementationClass());
      
      Map<String, String> propsByField = new HashMap<>();
      Iterator<P2JIndexComponent> compIter = index.components();
      while (compIter.hasNext())
      {
         P2JIndexComponent comp = compIter.next();
         String fieldName = comp.getName();
         String propertyName = propsByField.get(fieldName);
         if (propertyName == null)
         {
            Iterator<Property> itrProps = pclass.getDeclaredPropertyIterator();
            while (itrProps.hasNext())
            {
               Property hprop = itrProps.next();
               if (hprop.getColumnSpan() == 1)
               {
                  String field = ((Column) hprop.getColumnIterator().next()).getName();
                  if (field.equals(fieldName))
                  {
                     propertyName = hprop.getName();
                     break;
                  }
               }
               else
               {
                  // TODO: multi-col span properties
               }
            }
            
            propsByField.put(fieldName, propertyName);
         }
         
         String histName = TableMapper.getLegacyFieldName(buf, propertyName);
         if (histName != null)
         {
            // include only legacy fields
            root.addChild(buildIndexComponent(comp, propMap.get(normalizeName(histName))));
         }
      }
      
      return root;
   }
   
   /**
    * Build an {@link DataModelAst index component} associated with the given property AST.
    * 
    * @param    comp
    *           The index component
    * @param    propAst
    *           The associated {@link DataModelAst property AST}.
    * 
    * @return   See above.
    */
   private static DataModelAst buildIndexComponent(P2JIndexComponent comp, DataModelAst propAst)
   {
      String name        = comp.getName();
      String historical  = (String) propAst.getAnnotation(P2OLookup.HISTORICAL);
      String datatype    = (String) propAst.getAnnotation(P2OLookup.DATATYPE);
      
      DataModelAst root = createDataModelAst(DataModelTokenTypes.INDEX_COL, name.intern());
      root.putAnnotation(P2OLookup.HISTORICAL, historical.intern());
      root.putAnnotation(P2OLookup.DATATYPE, datatype.intern());
      if (comp.isIgnoreCase())
      {
         root.putAnnotation(P2OLookup.IGNORE_CASE, comp.isIgnoreCase());
      }
      if (comp.isDescending())
      {
         root.putAnnotation(P2OLookup.DESCEND, comp.isDescending());
      }
      if (comp.isAbbreviated())
      {
         root.putAnnotation(P2OLookup.ABBREVIATED, comp.isAbbreviated());
      }
      root.putAnnotation(P2OLookup.REFID, propAst.getId());
      
      return root;
   }
   
   /**
    * Create a {@link DataModelAst} with the given token type and text.
    * 
    * @param    type
    *           The node's token type.
    * @param    text
    *           The node's text.
    * 
    * @return   A {@link DataModelAst} instance setup as noted above.
    */
   private static DataModelAst createDataModelAst(int type, String text)
   {
      long treeId = AstManager.get().getTreeId(locate().p2oFile);
      DataModelAst ast = new DataModelAst(new CommonToken(type, text.intern()));
      ast.setId(AstManager.get().getNextNodeId(treeId));
      
      return ast;
   }
   
   /**
    * Create a {@link ProgressAst} with the given token type and text.
    * 
    * @param    type
    *           The node's token type.
    * @param    text
    *           The node's text.
    *           
    * @return   A {@link ProgressAst} instance setup as noted above.
    */
   private static ProgressAst createProgressAst(int type, String text)
   {
      return new ProgressAst(new CommonToken(type, text));
   }
   
   /**
    * Locate the context-local {@link WorkArea} instance.
    * 
    * @return   See above.
    */
   private static WorkArea locate()
   {
      return local.get();
   }
   
   /**
    * Shows a conversion error in a modal dialog and registers it in the error message list; does
    * not set the ERROR-STATUS:ERROR flag. The format of the message is:
    * <pre>
    * Could not [details] the '{@link WorkArea#predicate}' predicate for query
    * {@link WorkArea#qname}
    * </pre>
    * This method does the best effort to match the Progress message by analyzing the failing token.
    *
    * @param   exc
    *          The exception causing the conversion error.
    * @param   details
    *          The type of operation that failed.
    * @param   pCode
    *          The progress code being processed.
    */
   private static void showConversionError(Exception exc, String details, String pCode)
   {
      String msg = null;
      WorkArea wa = locate();
      
      if (exc instanceof NoViableAltException)
      {
         boolean isFind = pCode.startsWith(FIND_FIRST_TOK);
         Token tok = ((NoViableAltException) exc).token;
         
         if (tok.getType() == ProgressParserTokenTypes.DB_SYMBOL)
         {
            // TODO: try to detect if error 7325: each buffer must be added to dynamic query:
            // 'QUERY-PREPARE text must have 1 FOR EACH/PRESELECT for each query buffer'
            
            msg = tok.getText() + 
                  " must be a quoted constant or an unabbreviated, unambiguous\n" +
                  "buffer/field reference for buffers known to query ";
            if (isFind) 
            {
               msg += "or FIND"; 
            }
            ErrorManager.recordOrShowError(7328, msg, true, false, false);
         }
         else
         {
            int[] errCodes = new int[2];
            String[] errMsgs = new String[2];
            
            int errLoc = tok.getColumn() - 2; // skip a space to previous token
            int phraseStart = errLoc - 12;    // TODO: this is also heuristic, unknown steps back
            while (phraseStart > 0 && pCode.charAt(phraseStart) != ' ') 
            {
               // make sure we display the entire token
               phraseStart--;
            }
            // find the original predicate substring start 
            int predicateStart = pCode.indexOf(wa.predicate);
            if (predicateStart == -1)
            {
               phraseStart = 0; // should never happen because the pCode is built on wa.predicate
            }
            // 'drop' back any prefixes we added ro predicate to form a P4GL statement
            if (phraseStart < predicateStart)
            {
               phraseStart = predicateStart;
            }
            String phrase = pCode.substring(phraseStart, errLoc).trim();
            if (phrase.isEmpty()) 
            {
               // this will happen when the 1st token of the wa.predicate is faulty, because there
               // are no tokens before to show, we display it  
               phrase = tok.getText();
            }
            errCodes[0] = 247;
            errMsgs[0] = "** Unable to understand after -- \"" + phrase + "\"";
            msg = errMsgs[0]; 
            
            if (isFind)
            {
               errCodes[1] = 10086;
               errMsgs[1] = "FIND METHOD syntax is: [WHERE []] [USE-INDEX ]";
            }
            else
            {
               errCodes[1] = 7324;
               errMsgs[1] = "PREPARE syntax is: {FOR | PRESELECT} EACH  OF.. WHERE ... etc\"";
            }
            ErrorManager.recordOrShowError(errCodes, errMsgs, true, false, false);
         }
      }
      else
      {
         msg = "Could not %s the '%s' predicate for query %s";
         msg = String.format(msg, details, wa.predicate, wa.qname);
         ErrorManager.recordOrShowError(-1, msg, true, false, false);
      }
      
      if (LOG.isLoggable(Level.SEVERE))
      {
         LOG.log(Level.SEVERE, msg, exc);
      }
   }
   
   /**
    * Container of context-local data.
    * <p>
    * Implements {@link Finalizable} to unregister any surviving dynamic queries from the
    * class-loader.
    */
   private static class WorkArea
   {
      // temporary data about the currently converted query
      
      /** The query predicate being converted. */
      private String predicate = null;
      
      /** The Id of the current query, automatically incremented. */
      long queryId = 0;
      
      /** The locking type for the current query. May be null */
      private LockType lock = null;
      
      /** The name of the QUERY resource. */
      private String qname = null;
      
      /** The buffers used by the QUERY resource. */
      private List<Buffer> buffers = null;
      
      /** The file name of the class containing the converted query code. */
      private String dynQueryFile = null;
      
      /** The name of the generated {@link JavaAst} tree. */
      private String jastFile = null;
      
      /** The name of the converted {@link ProgressAst} tree. */
      private String astFile = null;
      
      /** The name of the {@link DataModelAst} tree with the temp-table objects for this query. */
      private String p2oFile = null;
   }
   
   /**
    * Allows to customize the processing of a predicate for generation of the dynamic query.
    */
   private interface QueryProcessor
   {
      /**
       * Returns the name of the variable that will contain the result of the evaluation.
       * 
       * @return  the expected variable name.
       */
      public String getExpectedQueryName();
      
      /**
       * Creates a Progress syntax correct statement that uses the given predicate. It will be
       * later converted and compiled.
       *
       * @param   predicate
       *          The query predicate that needs to be dynamically executed.
       *
       * @return  A string with a Progress query statement that uses the predicate.
       */
      public String preparePredicate(String predicate);
      
      /**
       * Post-process the intermediary AST to clean it up from unneeded nodes that were
       * automatically created during the progress conversion and add accessor for the query.
       *
       * @param   jcode
       *          The intermediary jast as it was generated form prepared predicate.
       *
       * @throws  ConfigurationException
       *          if any error occurs loading the specified configuration profile.
       * @throws  AstException
       *          if any error occurs loading a persisted AST.
       */
      public void postprocessJavaAst(JavaAst jcode)
      throws ConfigurationException,
             AstException;
   }
   
   /**
    * Class that keys the already generated JAST trees within the {@code LFUAgingCache} of
    * {@code DynamicQueryHelper}.
    */
   private static class QueryCacheKey
   {
      /** The query that generated a JAST tree. */
      private String query;
      
      /** The DMO classes of the buffers  */
      private Class[] bufferDMOs;
      
      /** 
       * The names of the buffers
       * TODO: Normally this should be removed, or better said, replaced with a buffer matching
       * algorithm because there are the same queries, only the name of the buffer differ.
       */
      private String[] bufferNames;
      
      public QueryCacheKey(String query, List<Buffer> buffers)
      {
         this.query = query;
         bufferDMOs = new Class[buffers.size()];
         bufferNames = new String[buffers.size()];
         int i = 0;
         for (Buffer buf : buffers)
         {
            RecordBuffer rbuf = ((BufferImpl) buf).buffer();
            bufferDMOs[i] = rbuf.getDMOImplementationClass();
            bufferNames[i] = rbuf.getDMOAlias();
            // System.out.println("Buff: name=" + rbuf.getDMOAlias() + ":" + bufferDMOs[i]);
            i++;
         }
      }
      
      /**
       * Checks whether some other object is "equal to" this one. Both query string and the list
       * buffer of buffers must be equals.
       * 
       * @param   o
       *          Another object to compare to.
       *          
       * @return  {@code true} if this object is the same as the obj argument.
       */
      @Override
      public boolean equals(Object o)
      {
         if (this == o)
         {
            return true;
         }
         
         if (!(o instanceof QueryCacheKey))
         {
            // objects are only used as keys in the cache so the o will also be of same class
            return false;
         }
         QueryCacheKey that = (QueryCacheKey) o;
         
         return query.equals(that.query)                    &&
                Arrays.equals(bufferDMOs, that.bufferDMOs)  &&
                Arrays.equals(bufferNames, that.bufferNames);
      }
      
      /**
       * Computes a hash code value for the object. This method is supported for the benefit of
       * hash tables such as those provided by {@link java.util.HashMap}.
       * 
       * @return  an integer value based on the query string and the buffer involved.
       */
      @Override
      public int hashCode()
      {
         int result = query.hashCode();
         result = 31 * result + Arrays.hashCode(bufferNames) + Arrays.hashCode(bufferDMOs);
         return result;
      }
      
      @Override
      public String toString() {
         return "QueryCacheKey{" +
                "query='" + query + '\'' +
                ", bufferNames=" + Arrays.toString(bufferNames) +
                ", bufferDMOs=" + Arrays.toString(bufferDMOs) +
                '}';
      }
   }
}
