/*
** Module   : DynamicTablesHelper.java
** Abstract : Helper class for dynamic temporary tables.
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20130624 Created initial version.
** 002 SVL 20130822 Continued implementation.
** 003 SVL 20131115 Implemented tableExists, some refactoring for dynamic buffers registration
**                  and clean up.
** 004 CA  20131013 Added no-op deleted() method, required by the changes in Finalizable interface.
** 005 OM  20140117 Added support for runtime indexed access to fields from buffer.
** 006 SVL 20140123 Implemented get4GLTableName.
** 007 SVL 20140124 Refactored DMO class cleanup.
** 008 OM  20140123 Added legacy table name in createDynamicDMO. Also the fields ar normalized.
** 009 OM  20140128 Fixed historical annotation so the table/fields can be find by P2OLookup.
**                  Removed ConcurrentModificationException in cleanup.
** 010 ECF 20140226 Renamed generateDMOAst to generateSchemaAst. Added index ASTs to the schema
**                  AST produced by this method. Replaced Apache commons logging with J2SE
**                  logging.
** 011 ECF 20140404 Reimplemented dynamic conversion to use ConversionPool.
** 012 OM  20140414 Added case-sensitive field support for indexed character columns.
** 013 ECF 20140611 Fixed regression caused by 012; avoids remapping all character computed
**                  column information for every dynamic temp-table. DatabaseManager now handles
**                  this mapping within its registerDynamicTable method.
** 014 ECF 20140922 Reduced memory leak and increased performance by caching dynamic temp-table
**                  resources. This is an interim update, but the cleanup of expired cache
**                  resources currently is broken, in that DMO resources can be released under
**                  some circumstances while a Hibernate session is still using them. The
**                  temporary workaround is to define a high cache limit, so that the cache does
**                  not expire elements.
** 015 ECF 20150109 Preserve original case for legacy field names.
** 016 VMN 20150122 Fixes to use non-normalized legacy field and table names in annotations.
** 017 ECF 20150327 Added p2o_post processing step.
** 018 ECF 20150414 Replaced in-memory compiling of dynamically generated DMO source code with
**                  direct-to-bytecode assembly from Java ASTs.
** 019 ECF 20150430 Fixed memory leak of active temp-table builders in DMOResourceManager inner
**                  class. A proper clean-up mechanism for dynamic temp-tables which are no longer
**                  in use and have expired from the cache is still needed.
*/

package com.goldencode.p2j.persist;

import antlr.*;
import com.goldencode.asm.*;
import com.goldencode.ast.*;
import com.goldencode.cache.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.xml.*;
import org.hibernate.cfg.*;
import org.hibernate.dialect.*;
import org.hibernate.internal.*;
import org.hibernate.mapping.*;
import org.w3c.dom.*;
import java.io.*;
import java.util.*;
import java.util.logging.*;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.*;

/**
 * Helper class for dynamic temporary tables.
 */
class DynamicTablesHelper
{
   /** Base package for generated DMOs. */
   private static final String DMO_BASE_PACKAGE = "com.goldencode.p2j.persist.dynamic";
   
   /** Category name for DMO implementation classes, used in conversion rules. */
   private static final String DMO_IMPL_CATEGORY = "impl";
   
   /** Category name for DMO interfaces, used in conversion rules. */
   private static final String DMO_IFACE_CATEGORY = "iface";
   
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(DynamicTablesHelper.class.getName());
   
   /** Singleton instance. */
   private static volatile DynamicTablesHelper instance;
   
   /** Param type to Progress type. */
   private static final Map<ParmType, Integer> propTypeToProgress = new HashMap<>();
   
   /** In-memory class loader */
   private static AsmClassLoader systemInMemoryClassLoader;
   
   /** Cache for dynamically converted and compiled DMO resources */ // TODO: make configurable
   private final ExpiryCache<CacheKey, CacheData> cache = new LFUAgingCache<>("dyn_dmo", 4096);
   
   /** Lock object on which to synchronize cache operations */
   private final Object cacheLock = new Object();
   
   /** Containers for context-specific data. */
   private final ContextLocal<Context> context = new ContextLocal<Context>()
   {
      public WeightFactor getWeight() { return WeightFactor.LEVEL_8; };
      public Context initialValue() { return new Context(); }
      protected void cleanup(Context ctx) { ctx.cleanup(); }
   };
   
   /** Counter used for generation of unique names for dynamic tables. */
   private final AtomicLong tableCounter = new AtomicLong(1);
   
   /** Shared Hibernate <code>Configuration</code> object. */
   private final Configuration cfg =
      DatabaseManager.getConfigBySchema(DatabaseManager.TEMP_TABLE_SCHEMA);
   
   /** Cross-context manager which cleans up shared DMO resources once they go out of scope */
   private final DMOResourceManager dmoResourceManager = new DMOResourceManager();
   
   static
   {
      propTypeToProgress.put(ParmType.CHAR, ProgressParserTokenTypes.FIELD_CHAR);
      propTypeToProgress.put(ParmType.INT, ProgressParserTokenTypes.FIELD_INT);
      propTypeToProgress.put(ParmType.INT64, ProgressParserTokenTypes.FIELD_INT64);
      propTypeToProgress.put(ParmType.DEC, ProgressParserTokenTypes.FIELD_DEC);
      propTypeToProgress.put(ParmType.LOG, ProgressParserTokenTypes.FIELD_LOGICAL);
      propTypeToProgress.put(ParmType.DATE, ProgressParserTokenTypes.FIELD_DATE);
      propTypeToProgress.put(ParmType.DT, ProgressParserTokenTypes.FIELD_DATETIME);
      propTypeToProgress.put(ParmType.DTTZ, ProgressParserTokenTypes.FIELD_DATETIME_TZ);
      propTypeToProgress.put(ParmType.HANDLE, ProgressParserTokenTypes.FIELD_HANDLE);
      propTypeToProgress.put(ParmType.RECID, ProgressParserTokenTypes.FIELD_RECID);
      propTypeToProgress.put(ParmType.ROWID, ProgressParserTokenTypes.FIELD_ROWID);
      propTypeToProgress.put(ParmType.RAW, ProgressParserTokenTypes.FIELD_RAW);
      
      systemInMemoryClassLoader = AsmClassLoader.getInstance();
      if (systemInMemoryClassLoader == null)
      {
         throw new IllegalStateException("It is expected to have InMemoryClassLoader " +
                                         "in the chain of system class loaders.");
      }
   }
   
   /**
    * Private constructor to enforce singleton.
    */
   private DynamicTablesHelper()
   {
   }
   
   /**
    * Get singleton instance of the helper.
    *
    * @return  singleton instance of the helper.
    */
   static DynamicTablesHelper getInstance()
   {
      if (instance == null)
      {
         synchronized (DynamicTablesHelper.class)
         {
            if (instance == null)
            {
               instance = new DynamicTablesHelper();
               PatternEngine.setDebugLevel(ConversionDriver.MSG_NONE);
            }
         }
      }
      
      return instance;
   }
   
   /**
    * Get Hibernate field name by 4GL field name for a dynamic DMO.
    *
    * @param  dmoIface
    *         DMO interface that corresponds the target dynamic DMO class.
    * @param  name4gl
    *         4GL field name.
    *
    * @return Hibernate field name for the given DMO class and 4GL field name or <code>null</code>
    *         if no corresponding name was found.
    */
   static String getHibernateFieldName(Class<?> dmoIface, String name4gl)
   {
      DMOInfo dmoInfo = getInstance().context.get().getDMOInfo(dmoIface);
      if (dmoInfo == null)
      {
         return null;
      }
      
      return dmoInfo.legacyToHibernateFields.get(normalizeName(name4gl));
   }
   
   /**
    * Get Hibernate field name by field creation id for a dynamic DMO.
    *
    * @param  dmoIface
    *         DMO interface that corresponds the target dynamic DMO class.
    * @param  fieldId
    *         4GL field id.
    *
    * @return Hibernate field name for the given DMO class and 4GL field name or <code>null</code>
    *         if no corresponding name was found.
    */
   static String getHibernateFieldName(Class<?> dmoIface, int fieldId)
   {
      // the algorithm here is based on the way the hibernate field names are generated at
      // "conversion" time. See (construction of clsFieldsXXX mappings in createDynamicDMO())
      return String.format("field%d", fieldId);
   }
   
   /**
    * Get 4GL field name by Hibernate field name for a dynamic DMO.
    *
    * @param  dmoIface
    *         DMO interface that corresponds the target dynamic DMO class.
    * @param  fieldNameHibernate
    *         Hibernate field name.
    *
    * @return 4GL field name for the given DMO class and Hibernate field name or <code>null</code>
    *         if no corresponding name was found.
    */
   static String get4GLFieldName(Class<?> dmoIface, String fieldNameHibernate)
   {
      DMOInfo dmoInfo = getInstance().context.get().getDMOInfo(dmoIface);
      if (dmoInfo == null)
      {
         return null;
      }
      
      return dmoInfo.hibernateToLegacyFields.get(fieldNameHibernate);
   }
   
   /**
    * Normalize 4GL field or index name: drop trailing spaces and convert to lower case. Equality
    * tests for field or index names should be done in their normalized form.
    *
    * @param  name
    *         Field or index name to normalize.
    *
    * @return Normalized form of the field or index name.
    */
   static String normalizeName(String name)
   {
      return rightTrim(name.toLowerCase());
   }
   
   /**
    * Get table name and database name from the string which represents full table name
    *
    * @param fullTableName
    *        The string which represents full table name. Can have two formats:
    *        "databaseName.tableName" or "tableName".
    *
    * @return String array with two elements. The first element is table name. The second element
    *         is database name or <code>null</code> if the database name wasn't specified in the
    *         full table name.
    */
   static String[] parseTableName(String fullTableName)
   {
      String tableName;
      String databaseName = null;
      
      int pos = rightTrim(fullTableName).indexOf('.');
      if (pos < 0)
      {
         tableName = rightTrim(fullTableName);
      }
      else
      {
         // TODO database name should be case-insensitive, but for now it is considered to be
         // case-sensitive
         databaseName = rightTrim(fullTableName.substring(0, pos));
         tableName = rightTrim(fullTableName.substring(pos + 1));
      }
      
      return new String[] { tableName, databaseName };
   }
   
   /**
    * Allows to determine if the buffer is for a dynamic table.
    *
    * @param  buffer
    *         Buffer to check.
    *
    * @return <code>true</code> if buffer is for a dynamic table.
    */
   static boolean isDynamicTableBuffer(RecordBuffer buffer)
   {
      TempTable tempTable = buffer.getParentTable();
      
      return (tempTable != null && tempTable._dynamic());
   }
   
   /**
    * Get 4GL name of the table which corresponds the given record buffer.
    *
    * @param   buffer
    *          Record buffer.
    *
    * @return  See above.
    */
   static String get4GLTableName(RecordBuffer buffer)
   {
      // dynamic temp-table case
      TempTableBuilder ttb = getTableInfo(buffer);
      if (ttb != null)
      {
         // dynamic table
         return ttb.name().toStringMessage();
      }
      
      // all other cases
      Class<?> dmoIface = buffer.getDMOInterface();
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      Class<?> dmoClass = DMOIndex.getImplementingClass(schema, dmoIface);
      
      LegacyTable legacyTable = dmoClass.getAnnotation(LegacyTable.class);
      return legacyTable.name();
   }
   
   /**
    * Returns the {@link TempTableBuilder} object that contains information about the 4GL
    * representation of the table or <code>null</code> if there is no corresponding dynamic table
    * for the specified record buffer.
    *
    * @param  buffer
    *         Record buffer that corresponds to the target dynamic table.
    *
    * @return See above.
    */
   static TempTableBuilder getTableInfo(RecordBuffer buffer)
   {
      return (TempTableBuilder) buffer.getParentTable();
   }
   
   /**
    * Determines if the specified table exists.
    *
    * @param  persistence
    *         Persistence which corresponds the target database.
    * @param  tableName4GL
    *         4GL table name.
    *
    * @return <code>true</code> if the specified table exists.
    */
   static boolean tableExists(Persistence persistence, String tableName4GL)
   {
      String normalizedTableName = normalizeName(tableName4GL);
      Database database = persistence.getDatabase();
      
      if (database.equals(DatabaseManager.TEMP_TABLE_DB))
      {
          return BufferManager.get().getStaticTempTableDepth(normalizedTableName) != -1;
      }
      else
      {
         String schema = DatabaseManager.getSchema(database);
         return TableMapper.getDMOClass(schema, normalizedTableName) != null;
      }
   }
   
   /**
    * Search for an existing table with the specified name.
    *
    * @param  fullTableName
    *         Table name in format "databaseName.tableName" or "tableName". In the latter case
    *         table search is performed in all permanent databases and then the temporary
    *         database. Dynamic tables are ignored.
    *
    * @return persistence which corresponds the target database or <code>null</code> if there is
    *         no such table.
    */
   static Persistence findExistingTable(String fullTableName)
   {
      String[] arr = parseTableName(fullTableName);
      String srcTableName = arr[0];
      String srcDatabaseName = arr[1];
      
      if (srcDatabaseName != null)
      {
         // database is explicitly specified
         if (!ConnectionManager.connected(srcDatabaseName).booleanValue())
         {
            return null;
         }
         
         String physicalName = ConnectionManager.pdbName(srcDatabaseName).getValue();
         Persistence persistence = PersistenceFactory.getInstance(physicalName);
         
         if (tableExists(persistence, srcTableName))
         {
            return persistence;
         }
         else
         {
            return null;
         }
      }
      else
      {
         // search through all databases, check permanent databases first
         long numDbs = ConnectionManager.numDBs().getValue();
         for (int i = 1; i <= numDbs; i++)
         {
            String databaseName = ConnectionManager.ldbName(i).getValue();
            if (!ConnectionManager.connected(databaseName).booleanValue())
            {
               continue;
            }
            
            String physicalName = ConnectionManager.pdbName(databaseName).getValue();
            Persistence persistence = PersistenceFactory.getInstance(physicalName);
            
            if (!tableExists(persistence, srcTableName))
            {
               continue;
            }
            
            return persistence;
         }
         
         // check temporary database
         Persistence persistence = PersistenceFactory.getInstance(DatabaseManager.TEMP_TABLE_DB);
         
         if (tableExists(persistence, srcTableName))
         {
            return persistence;
         }
         else
         {
            return null;
         }
      }
   }
   
   /**
    * Get original 4GL table name. Despite the fact that we specify 4GL table name as an input
    * parameter, this parameter can have different letter case or may have trailing spaces (this
    * form is allowed when specifying table name in 4GL code).
    *
    * @param  persistence
    *         Persistence which corresponds the target database.
    * @param  tableName4GL
    *         4GL table name (case-insensitive, may have trailing spaces).
    *
    * @return original 4GL table name.
    */
   static String get4GLTableName(Persistence persistence, String tableName4GL)
   {
      // TODO implement when table names in the original letter case will be preserved
      return normalizeName(tableName4GL);
   }
   
   /**
    * Convenience method to remove trailing spaces from the string.
    *
    * @param  str
    *         Target string.
    *
    * @return the target string without trailing spaces.
    */
   private static String rightTrim(String str)
   {
      return TextOps.rightTrim(str).toStringMessage();
   }
   
   /**
    * Generate DMO interface, implementation class and mapping document and register the new DMO
    * type.
    *
    * @param   builder
    *          Temporary table builder containing definitions of the target table.
    * @param   legacyTableName
    *          The legacy name of the table.
    *
    * @return  <code>true</code> on success.
    */
   boolean createDynamicDMO(TempTableBuilder builder, String legacyTableName)
   {
      Context local = context.get();
      
      Collection<P2JField> allFields = builder.getFields();
      Collection<P2JIndex> allIndexes = builder.getIndexes();
      
      // check the cache first
      CacheKey cacheKey = new CacheKey(allFields, allIndexes);
      CacheData cacheData = null;
      
      synchronized (cacheLock)
      {
         cacheData = cache.get(cacheKey);
      }
      
      String fullIfaceName;
      String fullImplClassName;
      String hibernateTableName;
      Document hbmDocument;
      Class<?> dmoIface;
      Class<?> dmoClass;
      DMOInfo dmoInfo = null;
      Map<String, Class<?>> classes = null;
      Map<String, String> clsFields4GLtoHibernate;
      Map<String, String> clsFieldsHibernateTo4GL;
      ArrayList<P2JIndex> indexesHibernate;
      
      try
      {
         if (cacheData == null)
         {
            indexesHibernate = new ArrayList<>();
            clsFields4GLtoHibernate = new HashMap<>();
            clsFieldsHibernateTo4GL = new HashMap<>();
            
            long tableId = tableCounter.getAndIncrement();
            hibernateTableName = String.format("dtt%d", tableId);
            
            String ifaceName = String.format("DynamicRecord%d", tableId);
            String implClassName = String.format("DynamicRecord%dImpl", tableId);
            
            long fieldId = 1;
            for (P2JField field : allFields)
            {
               String fieldName4GL = field.getName();
               String fieldName4GLNorm = normalizeName(fieldName4GL);
               String fieldNameHibernate = String.format("field%d", fieldId++);
               clsFields4GLtoHibernate.put(fieldName4GLNorm, fieldNameHibernate);
               clsFieldsHibernateTo4GL.put(fieldNameHibernate, fieldName4GL);
            }
            
            Aast root = instance.generateSchemaAst(builder,
                                                   ifaceName,
                                                   hibernateTableName,
                                                   legacyTableName,
                                                   clsFields4GLtoHibernate);
            
            try
            {
               ConversionPool.Results results;
               
               results = ConversionPool.runTask(ConversionProfile.P2O, root);
               root = (Aast) results.getStoredObject("p2o.prog");
               ConversionPool.runTask(ConversionProfile.P2O_POST, results, root);
               
               root = (Aast) results.getStoredObject("p2o.peer");
               root.putAnnotation("permanent", false);
               root.putAnnotation("package", DMO_BASE_PACKAGE);
               
               // generate .hbm.xml
               results = ConversionPool.runTask(ConversionProfile.HIBERNATE, root);
               XmlAst hbmXml = (XmlAst) results.getStoredObject(ifaceName);
               hbmDocument = XmlAst.astToDom(hbmXml);
               
               // generate interface and implementation Java ASTs
               results = ConversionPool.runTask(ConversionProfile.JAVA_DMO, root);
               Aast impl = (Aast) results.getStoredObject(DMO_IMPL_CATEGORY, ifaceName);
               Aast iface = (Aast) results.getStoredObject(DMO_IFACE_CATEGORY, ifaceName);
               
               fullIfaceName = String.format("%s.%s.%s",
                                             DMO_BASE_PACKAGE,
                                             DatabaseManager.TEMP_TABLE_SCHEMA,
                                             ifaceName);
               fullImplClassName = String.format("%s.%s.impl.%s",
                                                 DMO_BASE_PACKAGE,
                                                 DatabaseManager.TEMP_TABLE_SCHEMA,
                                                 implClassName);
               
               // generate bytecode from Java ASTs
               impl.putAnnotation("dmo-iface", fullIfaceName);
               results = ConversionPool.runTask(ConversionProfile.BREW_DMO_ASM, iface, impl);
               
               // even though we've already taken the hit of converting, give the cache one last
               // try, to minimize the chance of duplicate classes
               synchronized (cacheLock)
               {
                  cacheData = cache.get(cacheKey);
                  
                  if (cacheData == null)
                  {
                     // TODO: use a separate class loader for each table, so the classes can be
                     // garbage collected when no longer in use
                     AsmClassLoader loader = systemInMemoryClassLoader;
                     List<String> keys;
                     classes = new HashMap<>();
                     byte[] code;
                     
                     // load interfaces first
                     keys = (List<String>) results.getStoredObject("iface.keys");
                     for (String name : keys)
                     {
                        code = (byte[]) results.getStoredObject("ifaces", name);
                        // enable the following 2 lines for debugging
//                        FileOutputStream fos = new FileOutputStream(name + ".class");
//                        fos.write(code);
                        Class<?> clazz = loader.loadClass(name, code);
                        classes.put(name, clazz);
                     }
                     
                     // load classes next, since they are dependent upon interfaces
                     keys = (List<String>) results.getStoredObject("class.keys");
                     for (String name : keys)
                     {
                        code = (byte[]) results.getStoredObject("classes", name);
                        // enable the following 2 lines for debugging
//                        FileOutputStream fos = new FileOutputStream(name + ".class");
//                        fos.write(code);
                        Class<?> clazz = loader.loadClass(name, code);
                        classes.put(name, clazz);
                     }
                     
                     dmoInfo = new DMOInfo(classes.values(),
                                           clsFields4GLtoHibernate,
                                           clsFieldsHibernateTo4GL);
                     
                     CacheData cd = new CacheData(dmoInfo,
                                                  fullIfaceName,
                                                  fullImplClassName,
                                                  hibernateTableName,
                                                  hbmDocument,
                                                  indexesHibernate,
                                                  classes);
                     
                     cache.put(cacheKey, cd);
                  }
               }
            }
            finally
            {
               // clean up after TRPL rules in the AST registry
               String pkg = DMO_BASE_PACKAGE.replace('.', '/');
               String ifacePath = String.format("%s/%s/%s.java",
                                                pkg,
                                                DatabaseManager.TEMP_TABLE_SCHEMA,
                                                ifaceName);
               String implPath = String.format("%s/%s/impl/%s.java",
                                               pkg,
                                               DatabaseManager.TEMP_TABLE_SCHEMA,
                                               implClassName);
               AstManager astManager = AstManager.get();
               astManager.removeTree(ifacePath);
               astManager.removeTree(implPath);
            }
         }
         else
         {
            fullIfaceName = cacheData.getDMOIfaceName();
            fullImplClassName = cacheData.getDMOClassName();
            hibernateTableName = cacheData.getHibernateTableName();
            hbmDocument = cacheData.getHBMDocument();
            classes = cacheData.getClasses();
            dmoInfo = cacheData.getDMOInfo();
            indexesHibernate = cacheData.getIndexesHibernate();
            clsFields4GLtoHibernate = dmoInfo.getLegacyToHibernateFields();
            clsFieldsHibernateTo4GL = dmoInfo.getHibernateToLegacyFields();
         }
         
         dmoIface = classes.get(fullIfaceName);
         dmoClass = classes.get(fullImplClassName);
         PersistentClass persistentClass;
         
         synchronized (cfg)
         {
            // clean up any resources no longer in use
            dmoResourceManager.releasePendingResources();
            
            persistentClass = cfg.getClassMapping(dmoClass.getName());
         }
         
         if (persistentClass == null)  // not already mapped
         {
            // collect the list of character fields that will take part in indexes
            Set<String> indexedColumns = new HashSet<>();
            Set<String> caseSensitiveCols = new HashSet<>();
            for (P2JIndex index4GL : allIndexes)
            {
               Iterator<P2JIndexComponent> iter = index4GL.components();
               while (iter.hasNext())
               {
                  P2JIndexComponent comp4GL = iter.next();
                  // add character fields that are part of an index
                  if (comp4GL.isCharType())
                  {
                     final String dtFieldName = clsFields4GLtoHibernate.get(comp4GL.getName());
                     indexedColumns.add(dtFieldName);
                     if (!comp4GL.isIgnoreCase())
                     {
                        caseSensitiveCols.add(dtFieldName);
                     }
                  }
               }
            }
            
            MutableSessionFactoryImpl sessionFactory = (MutableSessionFactoryImpl)
               DatabaseManager.getSessionFactory(DatabaseManager.TEMP_TABLE_DB);
            
            synchronized (cfg)
            {
               // SVL: as far as I've checked Configuration object is not accessed by Hibernate
               // outside synchronized blocks in this class.  DBUtils.makeTypeArray (which also
               // uses Configuration object), accesses Configuration.typeResolver which remains
               // intact in the synchronized blocks.
               
               P2JDialect dialect =
                  (P2JDialect) DatabaseManager.getDialect(DatabaseManager.TEMP_TABLE_DB);
               ORMHandler.walkChildren(indexedColumns, caseSensitiveCols, hbmDocument, dialect);
               cfg.addDocument(hbmDocument);
               cfg.buildMappings();
               persistentClass = cfg.getClassMapping(dmoClass.getName());
               
               sessionFactory.add(persistentClass, cfg);
            }
            
            // collect index names generated by conversion, map them by legacy name
            Map<String, String> idxp2j = new HashMap<>();
            LegacyIndexes lindexes = dmoClass.getAnnotation(LegacyIndexes.class);
            if (lindexes != null)
            {
               for (LegacyIndex lidx : lindexes.value())
               {
                  idxp2j.put(lidx.name().toLowerCase(), lidx.converted());
               }
            }
            
            // translate 4GL indexes to Hibernate indexes
            for (P2JIndex index4GL : allIndexes)
            {
               String legacyIndexName = index4GL.getName().toLowerCase();
               String indexNameHibernate = String.format("idx__%s", idxp2j.get(legacyIndexName));
               P2JIndex indexHibernate = new P2JIndex(hibernateTableName,
                                                      indexNameHibernate,
                                                      index4GL.isUnique(),
                                                      index4GL.isPrimary(),
                                                      index4GL.isWord());
               Iterator<P2JIndexComponent> iter = index4GL.components();
               while (iter.hasNext())
               {
                  P2JIndexComponent comp4GL = iter.next();
                  String fieldName = clsFields4GLtoHibernate.get(normalizeName(comp4GL.getName()));
                  indexHibernate.addComponent(fieldName,
                                              comp4GL.isDescending(),
                                              comp4GL.isCharType(),
                                              comp4GL.isIgnoreCase(),
                                              comp4GL.isAbbreviated());
               }
               
               indexesHibernate.add(indexHibernate);
            }
            indexesHibernate.trimToSize();
         }
         
         if (local.registerDMOInfo(dmoIface, dmoInfo))
         {
            DMOIndex.registerDynamicTable(dmoIface, dmoClass, indexesHibernate);
            DatabaseManager.registerDynamicTable(dmoIface, persistentClass);
         }
         
         builder.setDmoIface(dmoIface);
         builder.setPersistentClass(persistentClass);
         
         local.registerBuilder(builder);
      }
      catch (Exception e)
      {
         LOG.log(Level.SEVERE, e.getMessage(), e);
         ErrorManager.displayError("Error while generating dynamic table: " + e.getMessage());
         
         return false;
      }
      
      return true;
   }
   
   /**
    * Register the dynamic buffer for a dynamic table. That will ensure buffer clean up if the
    * parent table is deleted.
    *
    * @param   builder
    *          Builder to be registered.
    * @param   buffer
    *          Dynamic buffer for a dynamic table to be registered.
    */
   void registerBufferForDynamicTable(TempTableBuilder builder, TemporaryBuffer buffer)
   {
      Context local = context.get();
      local.registerBuffer(builder, buffer);
   }
   
   /**
    * Deregister the given temp-table builder from the DMO resource manager.
    * 
    * @param   builder
    *          Builder to be deregistered.
    */
   void deregisterBuilder(TempTableBuilder builder)
   {
      dmoResourceManager.builderInactive(builder);
   }
   
   /**
    * Generate AST that represents the target table and has the same structure as .p2o files from
    * the conversion stage.
    *
    * @param  builder
    *         Temporary table builder containing definitions of the target table.
    * @param  ifaceName
    *         The name of the target DMO interface.
    * @param  tableName
    *         The name of the target SQL table.
    * @param  legacyTableName
    *         The lecagy (Progress) name of the table.
    * @param  fields4GLtoHibernate
    *         Map 4GL field names to Hibernate names.
    *
    * @return the root node of the generated AST.
    */
   private Aast generateSchemaAst(TempTableBuilder builder,
                                  String ifaceName,
                                  String tableName,
                                  String legacyTableName,
                                  Map<String, String> fields4GLtoHibernate)
   {
      Aast root = new ProgressAst(new CommonToken(ProgressParserTokenTypes.DATABASE,
                                                  DatabaseManager.TEMP_TABLE_SCHEMA));
      
      ProgressAst cls = new ProgressAst(new CommonToken(ProgressParserTokenTypes.TABLE,
                                                        tableName));
      cls.putAnnotation("class-name", ifaceName);
      cls.putAnnotation("table-name", tableName);
      cls.putAnnotation(P2OLookup.HISTORICAL, legacyTableName);
      root.addChild(cls);
      
      // add field nodes to table
      Map<String, Aast> fieldAsts = new HashMap<>();
      for (P2JField field : builder.getFields())
      {
         String fieldName4GL = normalizeName(field.getName());
         String fieldNameHibernate = fields4GLtoHibernate.get(fieldName4GL);
         ParmType type = field.getType();
         int progressType = propTypeToProgress.get(type);
         ProgressAst property = new ProgressAst(new CommonToken(progressType, fieldNameHibernate));
         fieldAsts.put(fieldName4GL, property);
         property.putAnnotation("format", field.getFormat());
         property.putAnnotation("label", field.getLabel());
         property.putAnnotation(P2OLookup.HISTORICAL, field.getName());
         if (field.isCaseSensitive())
         {
            property.putAnnotation("case-sensitive", field.isCaseSensitive());
         }
         
         long extent = field.getExtent();
         if (extent > 1)
         {
            property.putAnnotation("extent", extent);
         }
         
         cls.addChild(property);
         
         // add initial value node to field
         BaseDataType initial = field.getInitial();
         if (initial != null)
         {
            ProgressAst initialNode = new ProgressAst(
               new CommonToken(ProgressParserTokenTypes.KW_INIT, ""));
            property.addChild(initialNode);
            
            int valType;
            if (initial.isUnknown())
               valType = ProgressParserTokenTypes.UNKNOWN_VAL;
            else
            {
               switch (type)
               {
                  case CHAR:
                     valType = ProgressParserTokenTypes.STRING;
                     break;
                  case INT:
                     valType = ProgressParserTokenTypes.NUM_LITERAL;
                     break;
                  case DEC:
                     valType = ProgressParserTokenTypes.DEC_LITERAL;
                     break;
                  case LOG:
                     valType = initial.equals(new logical(true))
                               ? ProgressParserTokenTypes.BOOL_TRUE
                               : ProgressParserTokenTypes.BOOL_FALSE;
                     break;
                  case DATE:
                     valType = ProgressParserTokenTypes.DATE_LITERAL;
                     // TODO: what about builtins (today and now)? The above will not handle this
                     // correctly -- we already have a date object at this point and the builtin
                     // info is lost
                     break;
                  // TODO: datetime and datetimetz, other data types?
                  default:
                     throw new IllegalStateException();
               }
            }
            
            String valStr;
            switch (type)
            {
               case CHAR:
                  valStr = "\"" + initial.toStringMessage() + "\"";
                  break;
               default:
                  valStr = initial.toStringMessage();
            }
            
            ProgressAst initialVal = new ProgressAst(new CommonToken(valType, valStr));
            initialNode.addChild(initialVal);
         }
      }
      
      // add index nodes to table
      for (P2JIndex index : builder.getIndexes())
      {
         String indexName = normalizeName(index.getName());
         ProgressAst indexAst = new ProgressAst(new CommonToken(ProgressParserTokenTypes.INDEX,
                                                                indexName));
         cls.addChild(indexAst);
         
         if (index.isPrimary())
         {
            indexAst.putAnnotation(P2OLookup.PRIMARY, true);
         }
         
         if (index.isUnique())
         {
            indexAst.putAnnotation(P2OLookup.UNIQUE, true);
         }
         
         if (index.isWord())
         {
            indexAst.putAnnotation(P2OLookup.WORD, true);
         }
         
         // add index fields to index
         Iterator<P2JIndexComponent> compIter = index.components();
         while (compIter.hasNext())
         {
            P2JIndexComponent comp = compIter.next();
            String compName = normalizeName(comp.getName());
            ProgressAst compAst = new ProgressAst(
               new CommonToken(ProgressParserTokenTypes.INDEX_FIELD, compName));
            indexAst.addChild(compAst);
            
            if (comp.isDescending())
            {
               compAst.putAnnotation(P2OLookup.DESCEND, true);
            }
            
            // TODO: what about ABBREVIATED and CASE-SENSITIVE options (not handled as part of
            // TRPL schema fixups today)?
         }
      }
      
      assignIds(root, 1);
      
      // now assign each index field's refid to the corresponding field's ID 
      Aast indexAst = cls.getImmediateChild(ProgressParserTokenTypes.INDEX, null);
      while (indexAst != null)
      {
         Aast compAst = indexAst.getImmediateChild(ProgressParserTokenTypes.INDEX_FIELD, null);
         while (compAst != null)
         {
            String fieldName4GL = compAst.getText();
            Aast fieldAst = fieldAsts.get(fieldName4GL);
            compAst.putAnnotation("refid", fieldAst.getId());
            compAst = indexAst.getImmediateChild(ProgressParserTokenTypes.INDEX_FIELD, compAst);
         }
         
         indexAst = cls.getImmediateChild(ProgressParserTokenTypes.INDEX, indexAst);
      }
      
      return root;
   }
   
   /**
    * Traverse an AST and assign unique IDs to all of its nodes.
    *
    * @param  root
    *         The root node of the AST.
    * @param  startId
    *         First available id for assignment.
    *
    * @return new first available id for assignment.
    */
   private long assignIds(Aast root, long startId)
   {
      root.setId(startId++);
      
      int childCount = root.getNumImmediateChildren();
      if (childCount > 0)
      {
         for (int i = 0; i < childCount; i++)
         {
            Aast child = root.getChildAt(i);
            startId = assignIds(child, startId);
         }
      }
      return startId;
   }
   
   /**
    * Class which contains information about a dynamically define DMO, including the compiled
    * interface and implementation classes, and mappings between the legacy and converted fields.
    */
   private static class DMOInfo
   {
      /** All classes loaded to support this DMO */
      private final Collection<Class<?>> loadedClasses;
      
      /** Mapping of legacy, 4GL field names to converted, property names */
      private final Map<String, String> legacyToHibernateFields;
      
      /** Mapping of converted, property names to legacy, 4GL field names */
      private final Map<String, String> hibernateToLegacyFields;
      
      /**
       * Constructor.
       * 
       * @param   loadedClasses
       *          All classes loaded to support this DMO.
       * @param   legacyToHibernateFields
       *          Mapping of legacy, 4GL field names to converted, property names.
       * @param   hibernateToLegacyFields
       *          Mapping of converted, property names to legacy, 4GL field names.
       */
      DMOInfo(Collection<Class<?>> loadedClasses,
              Map<String, String> legacyToHibernateFields,
              Map<String, String> hibernateToLegacyFields)
      {
         this.loadedClasses = loadedClasses;
         this.legacyToHibernateFields = legacyToHibernateFields;
         this.hibernateToLegacyFields = hibernateToLegacyFields;
      }
      
      /**
       * Get all classes loaded to support this DMO.
       * 
       * @return  See above.
       */
      Collection<Class<?>> getLoadedClasses()
      {
         return loadedClasses;
      }
      
      /**
       * Get the mapping of legacy, 4GL field names to converted, property names.
       * 
       * @return  See above.
       */
      Map<String, String> getLegacyToHibernateFields()
      {
         return legacyToHibernateFields;
      }
      
      /**
       * Get the mapping of converted, property names to legacy, 4GL field names.
       * 
       * @return  See above.
       */
      Map<String, String> getHibernateToLegacyFields()
      {
         return hibernateToLegacyFields;
      }
   }
   
   /**
    * Class which contains information associated with a specific, {@link TempTableBuilder}
    * instance.
    */
   private static class BuilderInfo
   {
      /** Set of buffers associated with this temp-table builder */
      private final Set<TemporaryBuffer> buffers = new HashSet<>();
      
      /**
       * Associate a {@link TemporaryBuffer} with this object.
       * 
       * @param   buffer
       *          Temp-table buffer associated with the builder.
       */
      void registerBuffer(TemporaryBuffer buffer)
      {
         buffers.add(buffer);
      }
      
      void finish(boolean lastUser)
      {
         Iterator<TemporaryBuffer> iter = buffers.iterator();
         while (iter.hasNext())
         {
            TemporaryBuffer buffer = iter.next();
            boolean dropTable = lastUser && !iter.hasNext();
            buffer.closeMultiplexScope(dropTable);
            buffer.finished();
         }
      }
   }
   
   /**
    * A hashable key that encapsulates the characteristics which uniquely define the structure of
    * a dynamic temp-table; namely, its field and index definitions.
    */
   private static class CacheKey
   {
      /** List of field definitions for a dynamic temp-table */
      private final List<P2JField> fields;
      
      /** List of index definition wrappers for a dynamic temp-table */
      private final List<IndexKey> indexes;
      
      /**
       * Constructor.
       * 
       * @param   fields
       *          Collection of field definitions for a dynamic temp-table.
       * @param   indexes
       *          Collection of index definitions for a dynamic temp-table.
       */
      CacheKey(Collection<P2JField> fields, Collection<P2JIndex> indexes)
      {
         this.fields = new ArrayList<P2JField>(fields);
         this.indexes = new ArrayList<IndexKey>(indexes.size());
         for (P2JIndex next : indexes)
         {
            this.indexes.add(new IndexKey(next));
         }
      }
      
      /**
       * Use the field and index definitions in this key to calculate a hash code. This
       * implementation is consistent with this class' implementation of {@link #equals(Object)}.
       * 
       * @return  Hash code.
       */
      @Override
      public int hashCode()
      {
         int result = 17;
         result = 37 * result + fields.hashCode();
         result = 37 * result + indexes.hashCode();
         
         return result;
      }
      
      /**
       * Use the field and index definitions in this key to determine equivalence with the given
       * object. This implementation is consistent with this class' implementation of {@link
       * #hashCode()}.
       * 
       * @return  <code>true</code> if this object is equivalent with <code>o</code>, else
       *          <code>false</code>.
       */
      @Override
      public boolean equals(Object o)
      {
         if (this == o)
         {
            return true;
         }
         
         if (o instanceof CacheKey)
         {
            CacheKey that = (CacheKey) o;
            
            return (this.fields.equals(that.fields) && this.indexes.equals(that.indexes));
         }
         
         return false;
      }
      
      /**
       * Instances of this class are used to wrap instances of {@link P2JIndex}, for purposes of
       * producing hash codes suitable for using instances of the enclosing class as hash map
       * keys. This class provides stricter implementations of <code>hashCode</code> and
       * <code>equals</code> than those provided by <code>P2JIndex</code> itself.
       */
      private static class IndexKey
      {
         /** The wrapped instance of an index definition */
         private final P2JIndex index;
         
         /**
          * Constructor.
          * 
          * @param   index
          *          Index definition to be wrapped.
          */
         IndexKey(P2JIndex index)
         {
            this.index = index;
         }
         
         /**
          * Use the content of the wrapped index definition to produce a hash code. This is used
          * instead of {@link P2JIndex#hashCode()}, because the latter implementation of this
          * method is special-purpose, taking only the names of index components into account and
          * ignores other index metadata.
          * 
          * @return  Hash code.
          */
         @Override
         public int hashCode()
         {
            int result = 17;
            
            result = 37 * result + index.getName().hashCode();
            result = 37 * result + (index.isPrimary() ? 1 : 0);
            result = 37 * result + (index.isUnique() ? 1 : 0);
            result = 37 * result + (index.isWord() ? 1 : 0);
            
            Iterator<P2JIndexComponent> iter = index.components();
            while (iter.hasNext())
            {
               result = 37 * result + iter.next().hashCode();
            }
            
            return result;
         }
         
         /**
          * Overrides the default implementation to provide an implementation consistent with
          * {@link #hashCode()}.
          * 
          * @return  <code>true</code> if this object is equivalent with <code>o</code>, else
          *          <code>false</code>.
          */
         @Override
         public boolean equals(Object o)
         {
            if (this == o)
            {
               return true;
            }
            
            if (o instanceof IndexKey)
            {
               IndexKey that = (IndexKey) o;
               
               if (!this.index.getName().equals(that.index.getName()) ||
                   this.index.isPrimary() != that.index.isPrimary()   ||
                   this.index.isUnique() != that.index.isUnique()     ||
                   this.index.isWord() != that.index.isWord()         ||
                   this.index.size() != that.index.size())
               {
                  return false;
               }
               
               Iterator<P2JIndexComponent> iter = index.components();
               for (int i = 0; iter.hasNext(); i++)
               {
                  if (!iter.next().equals(that.index.getComponent(i)))
                  {
                     return false;
                  }
               }
               
               return true;
            }
            
            return false;
         }
      }
   }
   
   /**
    * DMO data to be cached for a dynamically defined temp-table.
    */
   private static class CacheData
   {
      /** DMO info object */
      private final DMOInfo dmoInfo;
      
      /** Fully qualified DMO interface name */
      private final String dmoIfaceName;
      
      /** Fully qualified DMO implementation class name */
      private final String dmoClassName;
      
      /** Database table name */
      private final String hibernateTableName;
      
      /** Hibernate ORM document */
      private final Document hbmDocument;
      
      /** List of index definitions using converted names */
      private final ArrayList<P2JIndex> indexesHibernate;
      
      /** Map of fully qualified class names to classes */
      private final Map<String, Class<?>> classes;
      
      /**
       * Constructor.
       * 
       * @param   dmoInfo
       *          DMO info object.
       * @param   dmoIfaceName
       *          Fully qualified DMO interface name.
       * @param   dmoClassName
       *          Fully qualified DMO implementation class name.
       * @param   hibernateTableName
       *          Database table name.
       * @param   hbmDocument
       *          Hibernate ORM document.
       * @param   indexesHibernate
       *          List of index definitions using converted names.
       * @param   classes
       *          Map of fully qualified class names to classes.
       */
      CacheData(DMOInfo dmoInfo,
                String dmoIfaceName,
                String dmoClassName,
                String hibernateTableName,
                Document hbmDocument,
                ArrayList<P2JIndex> indexesHibernate,
                Map<String, Class<?>> classes)
      {
         this.dmoInfo = dmoInfo;
         this.dmoIfaceName = dmoIfaceName;
         this.dmoClassName = dmoClassName;
         this.hibernateTableName = hibernateTableName;
         this.hbmDocument = hbmDocument;
         this.indexesHibernate = indexesHibernate;
         this.classes = classes;
      }
      
      /**
       * Get DMO information.
       * 
       * @return  See above.
       */
      DMOInfo getDMOInfo()
      {
         return dmoInfo;
      }
      
      /**
       * Get fully qualified DMO interface name.
       * 
       * @return  See above.
       */
      String getDMOIfaceName()
      {
         return dmoIfaceName;
      }
      
      /**
       * Get fully qualified DMO implementation class name.
       * 
       * @return  See above.
       */
      String getDMOClassName()
      {
         return dmoClassName;
      }
      
      /**
       * Get database table name.
       * 
       * @return  See above.
       */
      String getHibernateTableName()
      {
         return hibernateTableName;
      }
      
      /**
       * Get hibernate ORM document.
       * 
       * @return  See above.
       */
      Document getHBMDocument()
      {
         return hbmDocument;
      }
      
      /**
       * Get list of index definitions using converted names.
       * 
       * @return  See above.
       */
      ArrayList<P2JIndex> getIndexesHibernate()
      {
         return indexesHibernate;
      }
      
      /**
       * Get map of fully qualified class names to classes.
       * 
       * @return  See above.
       */
      Map<String, Class<?>> getClasses()
      {
         return Collections.unmodifiableMap(classes);
      }
   }
   
   /**
    * Context local information, used to clean up resources when the context ends and to respond
    * to queries about temp-table resources in use.
    * <p>
    * TODO: currently, the cleanup code is based on an incorrect assumption that the context's
    * state is reset more frequently than it actually is, AND, in some cases, we are cleaning up
    * resources prematurely, causing persistence errors when detached objects in an old Hibernate
    * session are reattached to a new one. The temporary workaround for this is a high cache
    * limit, to prevent the cache from expiring elements.
    */
   private class Context
   implements Finalizable
   {
      /** Map of DMO interfaces to maps of users of these interfaces (and  associated info) */
      private final Map<Class<?>, Map<TempTableBuilder, BuilderInfo>> dmoUsers = new HashMap<>();
      
      /** Map of DMO interfaces to DMO information objects */
      private final Map<Class<?>, DMOInfo> dmoMap = new HashMap<>();
      
      /** Temp-table builders which are no longer in use in this context */
      private List<TempTableBuilder> finishedBuilders = null;
      
      /**
       * Constructor. Registers this object as a global finalizable.
       */
      private Context()
      {
         TransactionManager.registerFinalizable(this, true);
      }
      
      /**
       * Clean up local context - drop tables and DMO definitions.
       */
      public void finished()
      {
         if (dmoUsers.isEmpty())
         {
            return;
         }
         
         if (finishedBuilders == null)
         {
            finishedBuilders = new ArrayList<>();
         }
         
         for (Map<TempTableBuilder, BuilderInfo> map : dmoUsers.values())
         {
            finishedBuilders.addAll(map.keySet());
         }
         
         for (TempTableBuilder builder : finishedBuilders)
         {
            Class<?> dmoIface = builder.getDMOInterface();
            Map<TempTableBuilder, BuilderInfo> users = dmoUsers.get(dmoIface);
            if (users == null)
            {
               continue;
            }
            
            BuilderInfo builderInfo = users.remove(builder);
            if (builderInfo == null)
            {
               continue;
            }
            
            boolean lastUser = users.isEmpty();
            builderInfo.finish(lastUser);
         }
      }
      
      /**
       * No-op.
       */
      public void deleted()
      {
      }
      
      /**
       * No-op.
       */
      public void iterate()
      {
      }
      
      /**
       * No-op.
       */
      public void retry()
      {
      }
      
      /**
       * Clean up DMO resources no longer needed by temp-table builders in this context.
       */
      private void cleanup()
      {
         if (finishedBuilders == null)
         {
            return;
         }
         
         // DMO interfaces which already have been cleaned up
         Set<Class<?>> processedDMOs = new HashSet<>();
         
         for (TempTableBuilder builder : finishedBuilders)
         {
            try
            {
               Class<?> dmoIface = builder.getDMOInterface();
               if (!processedDMOs.add(dmoIface))
               {
                  continue;
               }
               
               dmoUsers.remove(dmoIface);
               dmoMap.remove(dmoIface);
               PersistentClass persistentClass = builder.getPersistentClass();
               Class<?> dmoClass = persistentClass.getMappedClass();
               DMOIndex.deregisterDynamicTable(dmoIface, dmoClass);
               DatabaseManager.deregisterDynamicTable(dmoIface, persistentClass);
            }
            finally
            {
               dmoResourceManager.builderInactive(builder);
            }
         }
         
         finishedBuilders = null;
      }
      
      /**
       * Register a dynamic temp-table builder in this context.
       * 
       * @param   builder
       *          Temp-table builder to be registered.
       */
      private void registerBuilder(TempTableBuilder builder)
      {
         dmoResourceManager.builderActive(builder);
         
         Class<?> dmoIface = builder.getDMOInterface();
         Map<TempTableBuilder, BuilderInfo> users = dmoUsers.get(dmoIface);
         if (users == null)
         {
            users = new HashMap<>();
            dmoUsers.put(dmoIface, users);
         }
         
         BuilderInfo builderInfo = users.get(builder);
         if (builderInfo == null)
         {
            builderInfo = new BuilderInfo();
            users.put(builder, builderInfo);
         }
      }
      
      /**
       * Associate a record (temporary) buffer with the dynamic temp-table builder which defined
       * and created the dynamic temp-table backing the buffer.
       * 
       * @param   builder
       *          Temp-table builder.
       * @param   buffer
       *          Associated temporary buffer.
       */
      private void registerBuffer(TempTableBuilder builder, TemporaryBuffer buffer)
      {
         Class<?> dmoIface = builder.getDMOInterface();
         Map<TempTableBuilder, BuilderInfo> users = dmoUsers.get(dmoIface);
         BuilderInfo builderInfo = users.get(builder);
         builderInfo.registerBuffer(buffer);
      }
      
      /**
       * Get the DMO information object associated with the given DMO interface.
       * 
       * @param   dmoIface
       *          DMO interface.
       * 
       * @return  See above.
       */
      private DMOInfo getDMOInfo(Class<?> dmoIface)
      {
         return dmoMap.get(dmoIface);
      }
      
      /**
       * Associate a DMO information object with the given DMO interface.
       * 
       * @param   dmoIface
       *          DMO interface.
       * @param   dmoInfo
       *          DMO information object.
       * 
       * @return  <code>true</code> if <code>dmoIface</code> was not already associated with a
       *          different object, else <code>false</code>.
       */
      private boolean registerDMOInfo(Class<?> dmoIface, DMOInfo dmoInfo)
      {
         return (dmoMap.put(dmoIface, dmoInfo) == null);
      }
   }
   
   /**
    * An instance of this class manages resources requiring cleanup, across all contexts. This
    * object is a registered cache listener, receiving notifications when elements are expired
    * from the cache. These are tracked until the are no longer actively in use by any context,
    * then their associated resources are cleaned/reclaimed.
    * <p>
    * TODO: this cleanup currently is broken, as it can release resources that are still being
    * used by a Hibernate session. The temporary workaround is to define a high cache limit,
    * but this needs to be fixed.
    */
   private class DMOResourceManager
   implements CacheExpiryListener<CacheKey, CacheData>
   {
      /** Temp-table builders which are currently active across all contexts */
      private final Set<TempTableBuilder> activeBuilders = new HashSet<>();
      
      /** Data which has expired from the cache, but has not yet been released */
      private final Set<CacheData> expiredCacheData = new HashSet<>();
      
      /**
       * Constructor which registers with the dynamic DMO cache.
       */
      DMOResourceManager()
      {
         cache.addCacheExpiryListener(this);
      }
      
      /**
       * This method is invoked when the DMO cache has expired one or more elements. We use it to
       * know when a DMO's mapping can be deregistered with the Hibernate session factory and the
       * associated DMO interface and implementation class(es) can be released.
       * 
       * @param   event
       *          An object containing information about the event.
       */
      public void cacheExpired(CacheExpiryEvent<CacheKey, CacheData> event)
      {
         Map<CacheKey, CacheData> expired = event.getExpiredEntries();
         for (CacheData data : expired.values())
         {
            expiredCacheData.add(data);
         }
      }
      
      /**
       * Track the given temp-table builder as being in use.
       * 
       * @param   builder
       *          Temp-table builder to track.
       */
      void builderActive(TempTableBuilder builder)
      {
         synchronized (cacheLock)
         {
            activeBuilders.add(builder);
         }
      }
      
      /**
       * Stop tracking the given temp-table builder as being in use.
       * 
       * @param   builder
       *          Temp-table builder to stop tracking.
       */
      void builderInactive(TempTableBuilder builder)
      {
         synchronized (cacheLock)
         {
            activeBuilders.remove(builder);
         }
      }
      
      /**
       * Release resources which have expired from the dynamic DMO cache, which are no longer in
       * active use. This involves deregistering DMO types with Hibernate and releasing the DMO
       * interface and class(es) that have been compiled for a dynamic temp-table.
       * <p>
       * TODO: this cleanup currently is broken, as it can release resources that are still being
       * used by a Hibernate session. Also, if there are expired cache data, but they are still
       * pinned by active builders, this method could be expensive to execute every time a new
       * table is defined. Better to run it periodically in a dedicated thread. The temporary
       * workaround is to define a high cache limit, but this needs to be fixed.
       */
      void releasePendingResources()
      {
         synchronized (cacheLock)
         {
            if (expiredCacheData.isEmpty())
            {
               return;
            }
            
            // collect all expired data resources
            Map<Class<?>, CacheData> deadCacheData = new HashMap<>();
            for (CacheData data : expiredCacheData)
            {
               Map<String, Class<?>> classMap = data.getClasses();
               String dmoIfaceName = data.getDMOIfaceName();
               Class<?> dmoIface = classMap.get(dmoIfaceName);
               deadCacheData.put(dmoIface, data);
            }
            
            // collect all currently active DMO interfaces and PersistenceClass instances
            Map<Class<?>, PersistentClass> pinnedResources = new HashMap<>();
            for (TempTableBuilder ttb : activeBuilders)
            {
               pinnedResources.put(ttb.getDMOInterface(), ttb.getPersistentClass());
            }
            Set<Class<?>> pinnedDMOs = pinnedResources.keySet();
            
            // remove all the expired entries which still are pinned by active builders
            deadCacheData.keySet().removeAll(pinnedDMOs);
            
            MutableSessionFactoryImpl sessionFactory = (MutableSessionFactoryImpl)
               DatabaseManager.getSessionFactory(DatabaseManager.TEMP_TABLE_DB);
            for (CacheData data : deadCacheData.values())
            {
               // release Hibernate resources
               String dmoImplName = data.getDMOClassName();
               PersistentClass pc = cfg.getClassMapping(dmoImplName); // caller has cfg lock
               cfg.remove(pc);
               sessionFactory.remove(pc);
               
               // release class resources
               DMOInfo info = data.getDMOInfo();
               Collection<Class<?>> loadedClasses = info.getLoadedClasses();
               for (Class<?> clz : loadedClasses)
               {
//                  systemInMemoryClassLoader.deregister(clz.getName());
               }
               
               // remove this item from the set of expired entries requiring cleanup
               expiredCacheData.remove(data);
            }
         }
      }
   }
}
