/*
** Module   : FontTable.java
** Abstract : Implementation of the FONT-TABLE resource plus other misc font-related APIs.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20140701 Created initial version.
** 002 CA  20140711 Added legacy metrics support and font details for legacy font-tables.
** 003 MAG 20140827 Move common code to BaseSysTable class.
** 004 CA  20141220 Removed some confusing blanks in getLegacyTextMetrics logging messages.
** 005 CA  20150317 Added the standard font table, tracking of fixed-size fonts and font's max 
**                  width. 
** 006 CA  20150507 Fixed the default window title font.
** 007 CA  20150502 The resource ID needs to be kept at the resource, so it can be accessed after
**                  the resource is deleted.
*/

package com.goldencode.p2j.ui;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.*;

import javax.xml.parsers.*;

import org.w3c.dom.*;
import org.w3c.dom.Element;
import org.xml.sax.*;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.p2j.util.Utils.DirScope;
import com.goldencode.proxy.*;
import com.goldencode.util.*;

/**
 * Implementation of the FONT-TABLE system handle (the APIs from the {@link FontTableResource}
 * interface).  In addition, this implements:
 * <ul>
 *    <li>the <code>SYSTEM-DIALOG FONT</code> statement, via the {@link #chooseFont} API</li>
 *    <li>the <code>PUT-KEY-VALUE FONT ALL</code> statement, via the {@link #saveFontTable} API
 *    </li>
 *    <li>the <code>PUT-KEY-VALUE FONT font-num</code> statement, via the {@link #saveFont} API
 *    </li>
 * </ul>
 */
public class FontTable
extends BaseSysTable
{
   /** logger */
   private static final Logger LOG = LogHelper.getLogger(GenericFrame.class.getName());

   /** The name of the node containing the font-table. */
   public static final String FONT_TABLE = "font-table";

   /** The name of the node containing the system font-table. */
   public static final String SYSTEM_FONT_TABLE = "system-font-table";
   
   /** The name of the node containing the window title font - system font. */
   public static final String WINDOW_TITLE_FONT = "window-title-font";

   /** The name of the node containing the custom fonts (i.e. bitmap fonts, others). */
   public static final String CUSTOM_FONTS = "custom-fonts";
   
   /** The name of the node containing the default font. */
   public static final String DEFAULT_FONT = "default-font";

   /** The name of the node containing the default fixed font. */
   public static final String DEFAULT_FIXED_FONT = "default-fixed-font";
   
   /** The fallback font, in case the default fonts are not specified. */
   private static final String FALLBACK_DEFAULT_FONT = "System, size 10";
   
   /** A map containing the legacy font metrics. */
   private static Map<String, LegacyFontMetrics> legacyFontMetrics = new HashMap<>();

   /** A map containing the legacy text metrics. */
   private static Map<String, LegacyTextMetrics> legacyTextMetrics = new HashMap<>();

   /** A set containing the name of the fixed fonts. */
   private static Set<String> fixedFonts = new TreeSet<>();

   /** A map containing the custom fonts. */
   private static Map<String, byte[]> customFonts = new HashMap<>();

   /** A map of standard fonts, to which a key defaults, if not specified in the directory. */
   private static Map<String, String> standardFontTable = new HashMap<>();
   
   /** Context local instance of this class. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      { 
         return new WorkArea();
      };
   };

   /** initialize the standard font*/
   static
   {
      // initialize the font metrics
      String[] files = { "font-metrics.xml", "font-metrics-ext.xml" };
      for (int i = 0; i < files.length; i++)
      {
         loadFontMetrics(files[i]);
      }
      
      // initialize the text metrics
      loadTextMetrics("text-metrics.xml");
      
      // initialize the custom fonts
      loadCustomFonts();
      
      // load the standard font table
      // default fonts
      standardFontTable.put("default-font", "MS Sans Serif, size=8");
      standardFontTable.put("default-fixed-font", "Courier New, size=8");

      // standard fonts
      standardFontTable.put("font0", "Courier New, size=8");
      standardFontTable.put("font1", "MS Sans Serif, size=8");
      standardFontTable.put("font2", "Courier New, size=8");
      standardFontTable.put("font3", "Courier New, size=8");
      standardFontTable.put("font4", "MS Sans Serif, size=8");
      standardFontTable.put("font5", "MS Sans Serif, size=10");
      standardFontTable.put("font6", "MS Sans Serif, size=8, bold");
      standardFontTable.put("font7", "MS Sans Serif, size=8");

      // system fonts
      standardFontTable.put("window-title-font", "Tahoma, size 8, bold");
      
      // freeze the maps
      fixedFonts        = Collections.unmodifiableSet(fixedFonts);
      legacyFontMetrics = Collections.unmodifiableMap(legacyFontMetrics);
      legacyTextMetrics = Collections.unmodifiableMap(legacyTextMetrics);
      customFonts       = Collections.unmodifiableMap(customFonts);
      standardFontTable = Collections.unmodifiableMap(standardFontTable);
   }

   /**
    * Get a the instance for the FONT-TABLE system handle.  Is obtained using
    * a call to the {@link StaticProxy#obtain(Class, Class[])}, using the
    * {@link FontTableResource} interface and its methods implemented by these
    * classes: {@link FontTable}.
    * 
    * @return   See above.
    */
   public static handle asHandle()
   {
      WorkArea wa = local.get();

      Class<?>[] classes = { FontTable.class };
      FontTableResource proxy = StaticProxy.obtain(FontTableResource.class,
                                                   classes);

      if (wa.id == null)
      {
         wa.id = handle.resourceId(proxy);
      }

      return new handle(proxy);
   }   
   
   /**
    * Get this resource's ID.
    * 
    * @return   The resource's ID.
    */
   public static Long id()
   {
      WorkArea wa = local.get();
      
      return wa.id;
   }
   
   /**
    * Set this resource's ID.
    * <p>
    * This is a no-op for system handles.
    *  
    * @param    id
    *           The resource's ID.
    */
   public static void id(long id)
   {
      // no-op
   }

   /**
    * Implementation for the {@link WrappedResource#valid()} API.
    * 
    * @return   Always true.
    */
   public static boolean valid()
   {
      return true;
   }
   
   /**
    * Implementation for the {@link WrappedResource#unknown()} API.
    * 
    * @return   Always false.
    */
   public static boolean unknown()
   {
      return false;
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    *           
    * @see      handle#readOnlyError(handle, String)
    */
   public static void readOnlyError(String attribute)
   {
      handle.readOnlyError(asHandle(), attribute);
   }
   
   /**
    * Get the type of its associated handle.
    * 
    * @return   Always return the PSEUDO-WIDGET value.
    */
   public static character getResourceType()
   {
      return new character(LegacyResource.PSEUDO_WIDGET);
   }

   /**
    * Getter for the INSTANTIATING-PROCEDURE attribute.
    * <p>
    * The returned handle will refer a valid external procedure (persistent or not) or will be
    * unknown, in case its instantiating procedure was persistent and has been deleted.
    * 
    * @return   See above.
    */
   public static handle instantiatingProcedure()
   {
      // TODO: implement this
      throw new RuntimeException("INSTANTIATING-PROCEDURE attribute is not implemented.");
   }
   
   /**
    * Get the height in characters unit, of the default font.
    * 
    * @return   See above.
    */
   public static decimal getTextHeightChars()
   {
      return getTextHeightChars(new int64());
   }
   
   /**
    * Get the height in characters unit, of the specified font.  If the font number is unknown,
    * the default font is used.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public static decimal getTextHeightChars(long fontNum)
   {
      return getTextHeightChars(new int64(fontNum));
   }
   
   /**
    * Get the height in characters unit, of the specified font.  If the font number is unknown,
    * the default font is used.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public static decimal getTextHeightChars(int64 fontNum)
   {
      if (LogicalTerminal.isChui())
      {
         return new decimal(1);
      }

      int pheight = getTextHeightPixels(fontNum).intValue();
      double height = (double) pheight / LogicalTerminal.getClient().getPixelsPerRow();
      
      return new decimal(height, 2);
   }

   /**
    * Get the height in pixels, of the default font.
    * 
    * @return   See above.
    */
   public static integer getTextHeightPixels()
   {
      return getTextHeightPixels(new int64());
   }
   
   /**
    * Get the height in pixels, of the specified font.  If the font number is unknown,
    * the default font is used.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public static integer getTextHeightPixels(long fontNum)
   {
      return getTextHeightPixels(new int64(fontNum));
   }

   /**
    * Get the height in pixels, of the specified font.  If the font number is unknown,
    * the default font is used.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public static integer getTextHeightPixels(int64 fontNum)
   {
      if (LogicalTerminal.isChui())
      {
         return new integer(8);
      }
      
      int font = (fontNum.isUnknown() ? FontManager.DEFAULT_FONT : fontNum.intValue());
      
      return new integer(LogicalTerminal.getClient().getFontHeightPixels(font));
   }

   /**
    * Get the width of the specified text, in character units, using the default font.
    * 
    * @param    txt
    *           The text.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(String txt)
   {
      return getTextWidthChars(new character(txt), new int64());
   }

   /**
    * Get the width of the specified text, in character units, using the default font.
    * 
    * @param    txt
    *           The text.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(character txt)
   {
      return getTextWidthChars(txt, new int64());
   }

   /**
    * Get the width of the specified text, in character units, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(String txt, long fontNum)
   {
      return getTextWidthChars(new character(txt), new int64(fontNum));
   }

   /**
    * Get the width of the specified text, in character units, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(character txt, long fontNum)
   {
      return getTextWidthChars(txt, new int64(fontNum));
   }

   /**
    * Get the width of the specified text, in character units, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(String txt, int64 fontNum)
   {
      return getTextWidthChars(new character(txt), fontNum);
   }

   /**
    * Get the width of the specified text, in character units, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(character txt, int64 fontNum)
   {
      if (txt.isUnknown())
      {
         return new decimal(0);
      }
      
      if (LogicalTerminal.isChui())
      {
         return new decimal(TextOps.lengthOf(txt));
      }
      
      int pwidth = getTextWidthPixels(txt, fontNum).intValue();
      double width = (double) pwidth / (double) LogicalTerminal.getClient().getPixelsPerColumn();
      
      return new decimal(width, 2);
   }

   /**
    * Get the width of the specified text, in pixels, using the default font.
    * 
    * @param    txt
    *           The text.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(String txt)
   {
      return getTextWidthPixels(new character(txt), new int64());
   }

   /**
    * Get the width of the specified text, in pixels, using the default font.
    * 
    * @param    txt
    *           The text.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(character txt)
   {
      return getTextWidthPixels(txt, new int64());
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(String txt, long fontNum)
   {
      return getTextWidthPixels(new character(txt), new int64(fontNum));
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(character txt, long fontNum)
   {
      return getTextWidthPixels(txt, new int64(fontNum));
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(String txt, int64 fontNum)
   {
      return getTextWidthPixels(new character(txt), fontNum);
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(character txt, int64 fontNum)
   {
      if (txt.isUnknown())
      {
         return new integer(0);
      }
      
      if (LogicalTerminal.isChui())
      {
         return new integer(8 * TextOps.lengthOf(txt));
      }
      
      int font = (fontNum.isUnknown() ? FontManager.DEFAULT_FONT : fontNum.intValue());
      String stxt = txt.toStringMessage();
      ClientExports client = LogicalTerminal.getClient();
      
      LegacyTextMetrics ltm = legacyTextMetrics.get(stxt);
      String fontKey = client.getFontKey(font);
      int width = 0;
      
      if (ltm != null && fontKey != null)
      {
         LegacyFontMetrics lfm = ltm.legacyMetrics.get(fontKey);
         if (lfm != null)
         {
            // use the legacy width (in pixels)
            width = lfm.width;
         }
      }
      
      if (width == 0)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            final String msg = 
               "Could not determine width for text [%s] using font [%d] with key [%s].";
            LOG.warning(String.format(msg, stxt, font, fontKey));
         }

         width = client.getTextWidthPixels(stxt, font);
      }
      
      return new integer(width);
   }

   /**
    * Get the number of entries in the font-table.
    * 
    * @return   See above.
    */
   public static integer getNumEntries()
   {
      return new integer(LogicalTerminal.getClient().fontTableSize());
   }
   
   /**
    * Set the number of entries in the font-table.
    * 
    * @param    numEntries
    *           The new size of the font-table. 
    */
   public static void setNumEntries(long numEntries)
   {
      setNumEntries(new int64(numEntries));
   }
   
   /**
    * Set the number of entries in the font-table.
    * 
    * @param    numEntries
    *           The new size of the font-table. 
    */
   public static void setNumEntries(int64 numEntries)
   {
      int size = numEntries.isUnknown() ? 0 : numEntries.intValue();
      LogicalTerminal.getClient().adjustFontTableSize(size);
   }

   /**
    * Save the user's font-table to the environment.  This implements the PUT-KEY-VALUE FONT ALL
    * statement.
    */
   public static void saveFontTable()
   {
      // TODO: implement this
      throw new RuntimeException("PUT-KEY-VALUE FONT ALL statement is not implemented.");
   }
   
   /**
    * Save the the specified font, from the user's font-table, to the environment.  This 
    * implements the <code>PUT-KEY-VALUE FONT font-num</code> statement.
    * 
    * @param    fontNum
    *           The font-table entry to be saved.
    */
   public static void saveFont(long fontNum)
   {
      saveFont(new int64(fontNum));
   }
   
   /**
    * Save the the specified font, from the user's font-table, to the environment.  This 
    * implements the <code>PUT-KEY-VALUE FONT font-num</code> statement.
    * 
    * @param    fontNum
    *           The font-table entry to be saved.
    */
   public static void saveFont(int64 fontNum)
   {
      // TODO: implement this
      throw new RuntimeException("PUT-KEY-VALUE FONT <font-num> statement is not implemented.");
   }
   
   /**
    * Save the the specified font, from the user's font-table, to the environment.  This 
    * implements the <code>PUT-KEY-VALUE FONT font-num</code> statement.
    * 
    * @param    fontNum
    *           The font-table entry to be saved.
    */
   public static void chooseFont(long fontNum)
   {
      chooseFont(new int64(fontNum), LogicalTerminal.defaultWindow());
   }

   /**
    * Change the specified font, in the user's environment.
    * This implements the <code>SYSTEM-DIALOG FONT font-num</code> statement.
    * 
    * @param    fontNum
    *           The font-table entry to be changed.
    */
   public static void chooseFont(int64 fontNum)
   {
      chooseFont(fontNum, LogicalTerminal.defaultWindow());
   }
   
   /**
    * Save the the specified font, from the user's font-table, to the environment.  This 
    * implements the <code>PUT-KEY-VALUE FONT font-num</code> statement.
    * 
    * @param    fontNum
    *           The font-table entry to be saved.
    * @param    h
    *           The window handle to which the edited font-table belongs.
    */
   public static void chooseFont(long fontNum, handle h)
   {
      chooseFont(new int64(fontNum), h);
   }

   /**
    * Change the specified font, in the user's environment.
    * This implements the <code>SYSTEM-DIALOG FONT font-num</code> statement.
    * 
    * @param    fontNum
    *           The font-table entry to be changed.
    * @param    h
    *           The window handle to which the edited font-table belongs.
    */
   public static void chooseFont(int64 fontNum, handle h)
   {
      int font = (fontNum.isUnknown() ? 0 : fontNum.intValue());
      
      if (h.isUnknown())
      {
         final String msg = "Could not find widget handle for window from expression";
         ErrorManager.recordOrThrowError(3138, msg, false);
         
         return;
      }
      
      WrappedResource res = h.getResource();
      if (!(res instanceof WindowWidget))
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            final String msg = "Handle %d is referring resource %s, which is not a window!";
            LOG.severe(String.format(msg, h.getResourceId(), res.getClass().toString()));
         }
         
         return;
      }
      
      if (font <= 0 || font > getNumEntries().intValue())
      {
         // negative fonts seem to be allowed, but this corrupts the font-table!
         
         final String msg = "Invalid font number, %d, given to common dialog";
         ErrorManager.recordOrThrowError(3188, String.format(msg, font), false);
         
         return;
      }
      
      WindowWidget window = (WindowWidget) res;
      LogicalTerminal.getClient().chooseFont(font, window.getId());
   }

   /**
    * Read the default fonts.
    * 
    * @return   An array with the font details, with default font on index o and default fixed
    *           font on index 1.
    */
   static String[] readDefaultFonts()
   {
      String[] fonts = new String[2];
      
      String node = FONT_TABLE + "/" + DEFAULT_FONT;
      fonts[0] = Utils.getDirectoryNodeString(null, node, standardFontTable.get(DEFAULT_FONT),
                                              DirScope.BOTH, null);
      fonts[0] = safeFont(fonts[0]);
      
      node = FONT_TABLE + "/" + DEFAULT_FIXED_FONT;
      fonts[1] = Utils.getDirectoryNodeString(null, node, standardFontTable.get(DEFAULT_FIXED_FONT),
                                              DirScope.BOTH, null);
      fonts[1] = safeFont(fonts[1]);
      
      return fonts;
   }
   
   /**
    * Read the font-table with the name of the legacy system fonts.
    * <p>
    * On index 0, is the window title font.
    * TBD: other fonts.
    * 
    * @return   The font table.  As 4GL does, this is 0-indexed.
    */
   static String[] readSystemFontTable()
   {
      // TODO: other fonts
      
      String[] fonts = new String[1];
      
      String node = SYSTEM_FONT_TABLE + "/" + WINDOW_TITLE_FONT;
      fonts[0] = Utils.getDirectoryNodeString(null, node, standardFontTable.get(WINDOW_TITLE_FONT), 
                                              DirScope.BOTH, null);
      fonts[0] = safeFont(fonts[0]);

      return fonts;
   }
   
   /**
    * Read the entire font-table.
    * 
    * @return   The font table.  As 4GL does, this is 0-indexed.
    */
   static String[] readFontTable()
   {
      return readTable(FONT_TABLE, "fonts", "font", standardFontTable);
   }
   
   /**
    * Get the custom definition of this font.  If none available, return <code>null</code>.
    * 
    * @param    name
    *           The font name.
    *           
    * @return   A byte array with the font's definition, else <code>null</code>.
    */
   static byte[] getCustomFontDefinition(String name)
   {
      return customFonts.get(name.toLowerCase());
   }
   
   /**
    * Get the legacy metrics for the given font, using the font specified by the given key.
    * 
    * @param    key
    *           The font's key.
    *           
    * @return   The legacy metrics, with width on index 0 and height on index 1.
    */
   static int[] getLegacyFontMetrics(String key)
   {
      LegacyFontMetrics lfm = legacyFontMetrics.get(key);
      if (lfm == null)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Font [" + key + "] has no legacy metrics.");
         }

         return null;
      }
      
      return new int[] { lfm.width, lfm.maxwidth, lfm.height };
   }
   
   /**
    * Get the legacy metrics for the given text, using the font specified by the given key.
    * 
    * @param    text
    *           The text.
    * @param    key
    *           The font's key.
    *           
    * @return   The legacy metrics, with width on index 0 and height on index 1.
    */
   static int[] getLegacyTextMetrics(String text, String key)
   {
      LegacyTextMetrics ltm = legacyTextMetrics.get(text);
      if (ltm == null)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Text [" + text+ "] has no legacy metrics with font [" + key + "].");
         }

         return null;
      }
      
      LegacyFontMetrics lfm = ltm.legacyMetrics.get(key);
      if (lfm == null)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Text [" + text+ "] has no legacy metrics with font [" + key + "].");
         }

         return null;
      }
      
      return new int[] { lfm.width, lfm.height };
   }
   
   /**
    * Check if the given font face name is for a fixed font, registered in the {@link #fixedFonts}
    * set.
    * 
    * @param   name
    *          The font name.
    */
   static boolean isLegacyFixedFont(String name)
   {
      return name != null && fixedFonts.contains(name.toLowerCase());
   }
   
   /**
    * Load the custom fonts.
    */
   private static void loadCustomFonts()
   {
      DirectoryService ds = DirectoryService.getInstance();
      
      if (!ds.bind())
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("FontTable.loadCustmFonts - could not bind directory.");
         }
         
         return;
      }
      
      String path = Utils.findDirectoryNodePath(ds, CUSTOM_FONTS, "container", false);
      if (path == null)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.warning("No custom fonts are defined in the directory.");
         }

         return;
      }
      
      String[] fonts = ds.enumerateNodes(path);
      
      for (int i = 0; i < fonts.length; i++)
      {
         String fontPath = path + "/" + fonts[i];

         String font = ds.getNodeString(fontPath + "/font", "value");
         String file = ds.getNodeString(fontPath + "/file", "value");
         
         try
         {
            InputStream is = null;

            if (new File(file).exists())
            {
               // first try it as an external file
               is = new FileInputStream(new File(file));
            }
            else
            {
               // try load it from the jars
               is = ClassLoader.class.getResourceAsStream((file.startsWith("/") ? "" : "/") + file);
            }
            
            if (is == null)
            {
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.log(Level.SEVERE,
                          "Could not load the custom font [" + font + "] from file " + file);
               }
               
               continue;
            }
            
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            byte[] bytes = new byte[512];
            while (is.available() > 0)
            {
               int length = is.read(bytes);
               baos.write(bytes, 0, length);
            }
            
            bytes = baos.toByteArray();
            customFonts.put(font.toLowerCase(), bytes);
         }
         catch (IOException e)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE,
                       "Error loading the custom font [" + font + "] from file " + file,
                       e);
            }
         }
      }

      if (!ds.unbind())
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("FontTable.loadCustmFonts - could not unbind directory.");
         }
         
         return;
      }
   }
   
   /**
    * Load the text metrics from the specified file.
    * 
    * @param    file
    *           The XML file with the text metrics.
    */
   private static void loadTextMetrics(String file)
   {
      try
      {
         InputStream is = ClassLoader.class.getResourceAsStream("/" + file);
         if (is == null)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, "Could not load the legacy text metrics from file " + file);
            }

            return;
         }
         
         String systemID = generateSystemID();

         Document root = XmlHelper.parse(is, true, systemID);
         NodeList texts = root.getElementsByTagName("text");
         int len = texts.getLength();
         for (int i = 0; i < len; i++)
         {
            Element text = (Element) texts.item(i);
            
            String value = text.getAttribute("value");
            LegacyTextMetrics metrics = legacyTextMetrics.get(value);
            if (metrics == null)
            {
               metrics = new LegacyTextMetrics(value);
               legacyTextMetrics.put(value, metrics);
            }
            
            NodeList fonts = text.getElementsByTagName("font");
            loadMetricsWorker(fonts, metrics.legacyMetrics);
         }
      }
      catch (ParserConfigurationException | IOException | SAXException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Error loading the legacy text metrics from file " + file, e);
         }
      }
   }
   
   /**
    * Load the font metrics from the specified file.
    * 
    * @param    file
    *           The XML file with the text metrics.
    */
   private static void loadFontMetrics(String file)
   {
      try
      {
         InputStream is = ClassLoader.class.getResourceAsStream("/" + file);
         if (is == null)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, "Could not load the legacy font metrics from file " + file);
            }

            return;
         }
         
         String systemID = generateSystemID();
         
         Document root = XmlHelper.parse(is, true, systemID);
         NodeList fonts = root.getElementsByTagName("font");
         loadMetricsWorker(fonts, legacyFontMetrics);
         
         NodeList fixedFontNodes = root.getElementsByTagName("fixed-font");
         int len = fixedFontNodes.getLength();
         for (int i = 0; i < len; i++)
         {
            Element elFont = (Element) fixedFontNodes.item(i);

            String name = elFont.getAttribute("name");
            name = name.toLowerCase();
            fixedFonts.add(name);
         }
      }
      catch (ParserConfigurationException | IOException | SAXException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Error loading the legacy text metrics from file " + file, e);
         }
      }
   }
   
   /**
    * Load the font metrics from the specified DOM nodes and load them in the specified map.
    * 
    * @param    fonts
    *           The DOM nodes defining the font metrics.
    * @param    m
    *           The map where to load the metrics.
    */
   private static void loadMetricsWorker(NodeList fonts, Map<String, LegacyFontMetrics> m)
   {
      int len = fonts.getLength();
      for (int i = 0; i < len; i++)
      {
         // Get next element and reject if not required type.
         Element elFont = (Element) fonts.item(i);
         
         String name       = elFont.getAttribute("name");
         String ssize      = elFont.getAttribute("size");
         String sbold      = elFont.getAttribute("bold");
         String sitalic    = elFont.getAttribute("italic");
         String sunderline = elFont.getAttribute("underline");
         String savgwidth  = elFont.getAttribute("width");
         String smaxwidth  = elFont.getAttribute("max-width");
         String sheight    = elFont.getAttribute("height");
         
         int     size      = Integer.parseInt(ssize);
         boolean bold      = "true".equalsIgnoreCase(sbold);
         boolean italic    = "true".equalsIgnoreCase(sitalic);
         boolean underline = "true".equalsIgnoreCase(sunderline);
         int     avgwidth  = Integer.parseInt(savgwidth);
         int     maxwidth  = smaxwidth.length() == 0 ? -1 : Integer.parseInt(smaxwidth);
         int     height    = Integer.parseInt(sheight);
         
         String key = FontManager.getFontKey(name, size, bold, italic, underline);
         LegacyFont font = new LegacyFont(name, size, bold, italic, underline);
         LegacyFontMetrics lfm = new LegacyFontMetrics(font, avgwidth, maxwidth, height);
         
         m.put(key, lfm);
      }
   }
   
   /**
    * Generate a system ID to allow us to find the DTD.  This approach assumes that the DTD lives 
    * in the same jar file as this class.
    * 
    * @return    A system ID to be used for XML validation.
    */
   private static String generateSystemID()
   {
      String clsLoc = FontTable.class.getName().replace('.', '/') + ".class";
      URL clsURL = ClassLoader.getSystemResource(clsLoc);
      String path = clsURL.toString();
      String systemID = path.substring(0, path.indexOf(clsLoc));
      
      return systemID;
   }
   
   /**
    * If the given font is <code>null</code> or empty string, return the 
    * {@link #FALLBACK_DEFAULT_FONT}.  Otherwise, return the given font.
    * 
    * @return   See above.
    */
   private static String safeFont(String font)
   {
      return (font == null || font.trim().length() == 0 ? FALLBACK_DEFAULT_FONT : font);
   }
   
   /**
    * Structure with dDetails about a given font.
    */
   private static class LegacyFont
   {
      /** The font name. */
      private final String name;

      /** The font point size. */
      private final int size;
      
      /** Flag indicating if the font style is bold. */
      private final boolean bold;
      
      /** Flag indicating if the font style is italic. */
      private final boolean italic;
      
      /** Flag indicating if the font style is underline. */
      private final boolean underline;
      
      /**
       * Create a new font with the specified details.
       * 
       * @param    name
       *           The font name.
       * @param    size
       *           The font point size.
       * @param    bold
       *           Flag indicating if the font style is bold.
       * @param    italic
       *           Flag indicating if the font style is italic.
       * @param    underline
       *           Flag indicating if the font style is underline.
       */
      public LegacyFont(String name, int size, boolean bold, boolean italic, boolean underline)
      {
         this.name      = name;
         this.size      = size;
         this.bold      = bold;
         this.italic    = italic;
         this.underline = underline;
      }
   }

   /**
    * Structure containing the legacy metrics, for a certain font.  These metrics may be 
    * associated with the font itself or with a certain text.
    */
   private static class LegacyFontMetrics
   {
      /** The details about the legacy font. */
      private final LegacyFont font;
      
      /** The legacy width. */
      private final int width;
      
      /** The legacy max width of this font. */
      private final int maxwidth;
      
      /** The legacy height. */
      private final int height;
      
      /**
       * Instantiate new metrics details.
       * 
       * @param    font
       *           The associated font.
       * @param    width
       *           The legacy width.
       * @param    maxwidth
       *           The legacy max width of this font.
       * @param    height
       *           The legacy height.
       */
      public LegacyFontMetrics(LegacyFont font, int width, int maxwidth, int height)
      {
         this.font     = font;
         this.width    = width;
         this.height   = height;
         this.maxwidth = maxwidth;
      }
   }

   /**
    * Structure containing the legacy text metrics, for a certain text.
    */
   private static class LegacyTextMetrics
   {
      /** The text to which the metrics belong. */
      private final String text;
      
      /** A font-key to legacy metrics map. */
      private final Map<String, LegacyFontMetrics> legacyMetrics = new HashMap<>();
      
      /**
       * Instantiate a new text metrics.
       * 
       * @param    text
       *           The text.
       */
      public LegacyTextMetrics(String text)
      {
         this.text = text;
      }
   }
   
   /**
    * Container for context-local data. 
    */
   private static class WorkArea
   {
      /** The resource's ID. */
      private Long id = null;
   }
}
